"""Test Element class."""

import pytest
from .test_base import auto_setup_uwg, setup_open_matlab_ref, calculate_tolerance
from uwg import Element, Material


def test_element_init():
    """Test Element init method."""
    # Material: (thermalCond, volHeat = specific heat * density)
    concrete = Material(1.311, 836.8 * 2240, 'Concrete')
    gypsum = Material(0.16, 830.0 * 784.9, 'Gypsum')
    stucco = Material(0.6918, 837.0 * 1858.0, 'Stucco')

    # Mass wall for LargeOffce, Pst80, Zone 1A (Miami)
    thicknessLst = [0.0254, 0.0508, 0.0508, 0.0508, 0.0508, 0.0127]
    materialLst = [stucco, concrete, concrete, concrete, concrete, gypsum]
    wall = Element(albedo=0.08, emissivity=0.92, layer_thickness_lst=thicknessLst,
                   material_lst=materialLst, vegcoverage=0, t_init=293,
                   horizontal=False, name='MassWall')

    # test repr
    wall.__repr__()


def test_element_dict():
    """Test Element dict method."""

    # Init
    # Material: (thermalCond, volHeat = specific heat * density)
    concrete = Material(1.311, 836.8 * 2240, 'Concrete')
    gypsum = Material(0.16, 830.0 * 784.9, 'Gypsum')
    stucco = Material(0.6918, 837.0 * 1858.0, 'Stucco')

    # Mass wall for LargeOffce, Pst80, Zone 1A (Miami)
    thicknessLst = [0.0254, 0.0508, 0.0508, 0.0508, 0.0508, 0.0127]
    materialLst = [stucco, concrete, concrete, concrete, concrete, gypsum]
    wall = Element(albedo=0.08, emissivity=0.92, layer_thickness_lst=thicknessLst,
                   material_lst=materialLst, vegcoverage=0, t_init=293,
                   horizontal=False, name='MassWall')

    # make dict
    eldict = wall.to_dict()
    wall2 = Element.from_dict(eldict)

    assert wall.albedo == pytest.approx(wall2.albedo, abs=1e-10)
    assert wall.material_lst[0].thermalcond == \
           pytest.approx(wall2.material_lst[0].thermalcond, abs=1e-10)
    assert wall.emissivity == pytest.approx(wall2.emissivity, abs=1e-10)
    assert wall.layer_thickness_lst[1] == \
           pytest.approx(wall2.layer_thickness_lst[1], abs=1e-10)

    with pytest.raises(AssertionError):
        eldict['type'] = 'BemDef'
        Element.from_dict(eldict)


def test_SurfFlux_with_waterStorage_start():
    """Edge case: test element SurfFlux against matlab reference

    When waterStorage > 0.0. This has to be hardcoded b/c doesn't get used
    otherwise.
    """

    testuwg = auto_setup_uwg()
    testuwg.generate()

    # We subtract 30 days and 11 hours
    # New time: Jan 1, 1:00
    testuwg.simTime.nt -= (30 * 24 * 12 + 23 * 12 + 11)

    # turn rural road waterStorage to 1.
    testuwg.rural.waterStorage = 0.005  # .5cm thick film (from wgmax constant)

    # run simulation
    testuwg._hvac_autosize()
    testuwg.simulate()

    # check date
    assert testuwg.simTime.month == 1
    assert testuwg.simTime.day == 1
    assert testuwg.simTime.secDay == pytest.approx(300., abs=1e-15)

    # Check waterStorage
    assert 0.005 == pytest.approx(testuwg.rural.waterStorage, 1e-15)


def test_SurfFlux_with_waterStorage_middle():
    """ Edge case: test element SurfFlux against matlab reference.

    When waterStorage > 0.0. This has to be hardcoded b/c doesn't get used otherwise.
    """
    testuwg = auto_setup_uwg()
    testuwg.generate()

    # We subtract 30 days and 11 hours
    # New time: Jan 1, 1:00
    testuwg.simTime.nt -= (30 * 24 * 12 + 11 * 12)

    # turn rural road waterStorage to 1.
    testuwg.rural.waterStorage = 0.005  # .5cm thick film (from wgmax constant)

    # run simulation
    testuwg._hvac_autosize()
    testuwg.simulate()

    # check date
    assert testuwg.simTime.month == 1
    assert testuwg.simTime.day == 1
    assert testuwg.simTime.secDay/3600. == pytest.approx(13.0, abs=1e-15)

    # Check waterStorage
    assert 0.004643553960210 == pytest.approx(testuwg.rural.waterStorage, 1e-15)


def test_SurfFlux_unit():
    """Test element SurfFlux against matlab references at the start of timestep."""

    testuwg = auto_setup_uwg()

    testuwg._read_epw()
    testuwg.generate()

    # We subtract 23 hours and 55 minutes so we can test
    # initial timestep (1, 1, 300). New time: Jan 1, 5min
    testuwg.simTime.nt -= (23 * 12 + 24 * 12 * 30 + 11)

    # run simulation
    testuwg._hvac_autosize()
    testuwg.simulate()

    # check date is Jan 1st, 300s
    assert testuwg.simTime.month == 1
    assert testuwg.simTime.day == 1
    assert testuwg.simTime.secDay == pytest.approx(300.0, abs=1e-15)

    uwg_python_val = [
        testuwg.rural.aeroCond,        # Convection coef (refL uwg, eq.12)
        testuwg.rural.waterStorage,    # thickness of water film (m) (only for hor surf)
        testuwg.rural.solAbs,          # solar radiation absorbed (W m-2)
        testuwg.rural.lat,             # surface latent heat flux (W m-2)
        testuwg.rural.sens,            # surface sensible heat flux (W m-2)
        testuwg.rural.flux,            # external surface heat flux (W m-2)
        testuwg.rural.T_ext,           # external surface temperature (K)
        testuwg.rural.T_int]           # interior surface temperature (K)

    # Matlab Checking for rural road
    uwg_matlab_val = setup_open_matlab_ref(
        'matlab_element', 'matlab_ref_element_surfflux_winter.txt')

    # matlab ref checking
    assert len(uwg_matlab_val) == len(uwg_python_val)

    for i in range(len(uwg_matlab_val)):
        tol = calculate_tolerance(uwg_matlab_val[i], 15.0)
        assert uwg_python_val[i] == pytest.approx(uwg_matlab_val[i], abs=tol), \
            'error at index={}'.format(i)


def test_SurfFlux_integration():
    """Test element SurfFlux against matlab references at the end of timestep.

    Integration test as it requires other uwg classes to be working before functioning
    correctly.
    """

    testuwg = auto_setup_uwg()

    # Change time and vegCoverage parameters so we can get
    # effect of vegetation on surface heat flux
    testuwg.vegstart = 2    # February
    testuwg.nday = 31 + 15  # February, 15

    testuwg.generate()

    # We subtract 11 hours from total timestep so still have sun. New time: 1300
    testuwg.simTime.nt -= 12 * 11

    testuwg.simulate()

    # check date is February 15th, 13:00
    assert testuwg.simTime.month == 2
    assert testuwg.simTime.day == 15
    assert testuwg.simTime.secDay/3600. == pytest.approx(13.0, abs=1e-15)

    uwg_python_val = [
        testuwg.rural.aeroCond,        # Convection coef (refL uwg, eq.12)
        testuwg.rural.waterStorage,    # thickness of water film (m) (only for hor surf)
        testuwg.rural.solAbs,          # solar radiation absorbed (W m-2)
        testuwg.rural.lat,             # surface latent heat flux (W m-2)
        testuwg.rural.sens,            # surface sensible heat flux (W m-2)
        testuwg.rural.flux,            # external surface heat flux (W m-2)
        testuwg.rural.T_ext,           # external surface temperature (K)
        testuwg.rural.T_int]           # interior surface temperature (K)

    # Matlab Checking for rural road
    uwg_matlab_val = \
        setup_open_matlab_ref('matlab_element', 'matlab_ref_element_surfflux.txt')

    # matlab ref checking
    assert len(uwg_matlab_val) == len(uwg_python_val)

    for i in range(len(uwg_matlab_val)):
        tol = calculate_tolerance(uwg_matlab_val[i], 15.0)
        assert uwg_python_val[i] == pytest.approx(uwg_matlab_val[i], abs=tol), \
            'error at index={}'.format(i)



"""Tests for BEM object."""

# Fix for Python 2.X
try:
    unicode
except NameError:
    unicode = str

import pytest
from .test_base import auto_setup_uwg, setup_open_matlab_ref, calculate_tolerance
from uwg import BEMDef, Building, Element, Material


def _bemdef():
    """Create BEMDef: LargeOffce, Pst80, Zone 1A (Miami)."""

    # Material: (thermalCond, volHeat = specific heat * density)
    concrete = Material(1.311, 836.8 * 2240, 'Concrete')
    gypsum = Material(0.16, 830.0 * 784.9, 'Gypsum')
    stucco = Material(0.6918, 837.0 * 1858.0, 'Stucco')
    insulation = Material(0.049, 836.8 * 265.0, "Insulation")

    # Mass wall for LargeOffce, Pst80, Zone 1A (Miami)
    thicknessLst = [0.0254, 0.0508, 0.0508, 0.0508, 0.0508, 0.0127]
    materialLst = [stucco, concrete, concrete, concrete, concrete, gypsum]
    wall = Element(albedo=0.08, emissivity=0.92, layer_thickness_lst=thicknessLst,
                   material_lst=materialLst, vegcoverage=0, t_init=293,
                   horizontal=False, name='MassWall')

    # IEAD roof
    thicknessLst = [0.058, 0.058]
    materialLst = [insulation, insulation]
    roof = Element(albedo=0.2, emissivity=0.93, layer_thickness_lst=thicknessLst,
                   material_lst=materialLst, vegcoverage=0.5, t_init=293,
                   horizontal=True, name='IEAD')

    # Mass floor
    thicknessLst = [0.054, 0.054]
    materialLst = [concrete, concrete]
    floor = Element(albedo=0.2, emissivity=0.9, layer_thickness_lst=thicknessLst,
                    material_lst=materialLst, vegcoverage=0.0, t_init=293,
                    horizontal=True, name='MassFloor')

    bld = Building(floor_height=3.5, int_heat_night=1, int_heat_day=1, int_heat_frad=0.1,
                   int_heat_flat=0.1, infil=0.26, vent=0.0005, glazing_ratio=0.4,
                   u_value=5.8, shgc=0.2, condtype='AIR', cop=5.2,
                   coolcap=76, heateff=0.7, initial_temp=293)

    return BEMDef(bld, floor, wall, roof, bldtype='fullservicerestaurant',
                  builtera='pst80')


def test_bem_init_manual():
    """Test BEMDef init methods."""

    # test init
    bem = _bemdef()

    # test repr
    bem.__repr__()


def test_bemdef_dict():
    """Test BEMDef dict methods."""

    bem = _bemdef()

    # make dict
    bemdict = bem.to_dict()

    # test if dict and from_dict
    assert isinstance(bemdict, dict)

    bem2 = BEMDef.from_dict(bemdict)

    assert bem.wall.albedo == pytest.approx(bem2.wall.albedo, abs=1e-10)
    assert bem.frac == pytest.approx(bem2.frac, abs=1e-10)
    assert bem.building.cop == pytest.approx(bem2.building.cop, abs=1e-10)
    assert bem.mass.layerThermalCond[0] == \
        pytest.approx(bem2.mass.layerThermalCond[0], abs=1e-10)

    with pytest.raises(AssertionError):
        bemdict['type'] = 'BemDef'
        BEMDef.from_dict(bemdict)


def test_bem_building_init_largeoffice_readDOE():
    """Test for BEM.building init"""

    testuwg = auto_setup_uwg()
    testuwg.generate()

    uwg_python_val = [
        testuwg.BEM[0].building.floor_height,
        testuwg.BEM[0].building.int_heat,
        testuwg.BEM[0].building.int_heat_night,
        testuwg.BEM[0].building.int_heat_day,
        testuwg.BEM[0].building.int_heat_frad,
        testuwg.BEM[0].building.int_heat_flat,
        testuwg.BEM[0].building.infil,
        testuwg.BEM[0].building.vent,
        testuwg.BEM[0].building.glazing_ratio,
        testuwg.BEM[0].building.u_value,
        testuwg.BEM[0].building.shgc,
        testuwg.BEM[0].building.condtype,
        testuwg.BEM[0].building.cop,
        testuwg.BEM[0].building.cool_setpoint_day,
        testuwg.BEM[0].building.cool_setpoint_night,
        testuwg.BEM[0].building.heat_setpoint_day,
        testuwg.BEM[0].building.heat_setpoint_night,
        testuwg.BEM[0].building.coolcap,
        testuwg.BEM[0].building.heat_cap,
        testuwg.BEM[0].building.heateff,
        testuwg.BEM[0].building.msys,
        testuwg.BEM[0].building.indoor_temp,
        testuwg.BEM[0].building.indoor_hum,
        testuwg.BEM[0].building.FanMax,
        testuwg.BEM[0].bldtype,
        testuwg.BEM[0].builtera,
        testuwg.BEM[0].zonetype]

    uwg_matlab_val = setup_open_matlab_ref(
        'matlab_bem', 'matlab_ref_bem_building_init_largeoffice.txt')

    # matlab ref checking
    assert len(uwg_matlab_val) == len(uwg_python_val)
    for i in range(len(uwg_matlab_val)):
        if isinstance(uwg_python_val[i], (str, unicode)):
            assert ''.join(uwg_python_val[i].lower().split()) == \
                ''.join(uwg_matlab_val[i].lower().split()), \
                'error at index={}'.format(i)
        else:
            tol = calculate_tolerance(uwg_python_val[i], 14.0)
            assert uwg_python_val[i] == \
                pytest.approx(
                    uwg_matlab_val[i], abs=tol), 'error at index={}'.format(i)


def test_bem_building_bemcalc_largeoffice_cooling_readDOE():
    """Test for bem.building bemcalc during cooling period."""

    testuwg = auto_setup_uwg()

    # Test Jan 1 (winter, no vegetation coverage)
    testuwg.month = 1
    testuwg.day = 1
    testuwg.nday = 1

    # set_input
    testuwg.generate()

    # In order to avoid integration effects. Test only first time step
    # Subtract timestep to stop at 300 sec
    testuwg.simTime.nt -= (23 * 12 + 11)

    # Run simulation
    testuwg.simulate()

    # check date
    assert testuwg.simTime.month == 1
    assert testuwg.simTime.day == 1
    assert testuwg.simTime.secDay == pytest.approx(300.0, abs=1e-15)

    # Calculated values
    uwg_python_val = [
        testuwg.BEM[0].building.int_heat,
        testuwg.BEM[0].building.int_heat_night,
        testuwg.BEM[0].building.int_heat_day,
        testuwg.BEM[0].building.indoor_temp,
        testuwg.BEM[0].building.indoor_hum,
        testuwg.BEM[0].building.indoorRhum,
        testuwg.BEM[0].building.nFloor,
        testuwg.BEM[0].building.sensCoolDemand,
        testuwg.BEM[0].building.sensHeatDemand,
        testuwg.BEM[0].building.cop_adj,
        testuwg.BEM[0].building.dehumDemand,
        testuwg.BEM[0].building.coolConsump,
        testuwg.BEM[0].building.heatConsump,
        testuwg.BEM[0].building.sensWaste,
        testuwg.BEM[0].building.fluxMass,
        testuwg.BEM[0].building.fluxWall,
        testuwg.BEM[0].building.fluxRoof,
        testuwg.BEM[0].building.fluxSolar,
        testuwg.BEM[0].building.fluxWindow,
        testuwg.BEM[0].building.fluxInterior,
        testuwg.BEM[0].building.fluxInfil,
        testuwg.BEM[0].building.fluxVent,
        testuwg.BEM[0].building.ElecTotal,
        testuwg.BEM[0].building.GasTotal,
        testuwg.BEM[0].building.Qhvac,
        testuwg.BEM[0].building.Qheat]

    # Note: the following properties are instantiated but never set in UWG_Matlab
    # testuwg.BEM[0].building.latWaste
    # testuwg.BEM[0].building.RadFOcc
    # testuwg.BEM[0].building.LatFOcc
    # testuwg.BEM[0].building.RadFEquip
    # testuwg.BEM[0].building.RadFLight
    # testuwg.BEM[0].building.Tdp

    uwg_matlab_val = setup_open_matlab_ref(
        'matlab_bem', 'matlab_ref_bem_building_bemcalc_largeoffice_cooling.txt')

    # matlab ref checking
    assert len(uwg_matlab_val) == len(uwg_python_val)

    for i in range(len(uwg_matlab_val)):
        tol = calculate_tolerance(uwg_python_val[i], 15.0)
        assert uwg_python_val[i] == \
            pytest.approx(uwg_matlab_val[i],
                          abs=tol), 'error at index={}'.format(i)



"""Geographic Parameters class."""


class Param(object):
    """Geographic Parameters class.

    Args:
        dayBLHeight: Daytime mixing height [orig: 700] [m].
        nightBLHeight: Nighttime boundary-layer height
            (orig: 80, Sing: 80, Bub-Cap: 50) [m]
        refHeight: Reference height at which the vertical profile of potential
            temperature is vertical
        tempHeight: Temperature measuremnt height at the weather station [m]
        windHeight: Air velocity measuremnt height at the weather station [m]
        circCoeff: Wind scaling coefficient
        dayThreshold: Heat flux threshold for daytime conditions [W m-2]
        nightThreshold: Heat flux threshold for nighttime conditions [W m-2]
        treeFLat: Value between 0 and 1 for latent fraction of trees
        grassFLat: Value between 0 and 1 for latent fraction of grass
        vegAlbedo: Value between 0 and 1 for albedo of vegetation
        vegStart: Value between 1 and 12 for begin month for vegetation participation
        vegEnd: Value between 1 and 12 for end month for vegetation participation
        nightSetStart: Begin hour for night thermal set point schedule
        nightSetEnd: End hour for night thermal set point schedule
        windMin: Minimum wind speed [m s-1]
        wgmax: Maximum film water depth on horizontal surfaces [m]
        exCoeff: Exchange velocity coefficient
        maxdx: Maximum discretization length for the UBL model [m]
        g: Gravity [m s-2]
        cp: Heat capacity for air [J/kg K]
        vk: Von karman constant [dimensionless]
        r: Gas constant for dry air [J/kg K]
        rv: Gas constant for water vapor [J/kg K]
        lv: Latent heat of evaporation [J/kg]
        pi: Pi
        sigma: Stefan Boltzmann constant [W K-4 m-2]
        waterDens: Water density [kg m-3]
        lvtt: Undefined.
        tt: Undefined.
        estt: Undefined.
        cl: Undefined.
        cpv: Undefined.
        b  : Coefficients derived by Louis [1979]
        cm : Undefined.
        colburn: [Pr/Sc]^[2/3] for Colburn analogy in water evaporation

    Properties:
        * dayBLHeight
        * nightBLHeight
        * refHeight
        * tempHeight
        * windHeight
        * circCoeff
        * dayThreshold
        * nightThreshold
        * treeFLat
        * grassFLat
        * vegAlbedo
        * vegStart
        * vegEnd
        * nightSetStart
        * nightSetEnd
        * windMin
        * wgmax
        * exCoeff
        * maxdx
        * g
        * cp
        * vk
        * r
        * rv
        * lv
        * pi
        * sigma
        * waterDens
        * lvtt
        * tt
        * estt
        * cl
        * cpv
        * b
        * cm
        * colburn
    """

    def __init__(self, dayBLHeight, nightBLHeight, refHeight, tempHeight, windHeight,
                 circCoeff, dayThreshold, nightThreshold, treeFLat, grassFLat, vegAlbedo,
                 vegStart, vegEnd, nightSetStart, nightSetEnd, windMin, wgmax, exCoeff,
                 maxdx, g, cp, vk, r, rv, lv, pi, sigma, waterDens, lvtt, tt, estt, cl,
                 cpv, b, cm, colburn):

        self.dayBLHeight = dayBLHeight
        self.nightBLHeight = nightBLHeight
        self.refHeight = refHeight
        self.tempHeight = tempHeight
        self.windHeight = windHeight
        self.circCoeff = circCoeff
        self.dayThreshold = dayThreshold
        self.nightThreshold = nightThreshold
        self.treeFLat = treeFLat
        self.grassFLat = grassFLat
        self.vegAlbedo = vegAlbedo
        self.vegStart = vegStart
        self.vegEnd = vegEnd
        self.nightSetStart = nightSetStart
        self.nightSetEnd = nightSetEnd
        self.windMin = windMin
        self.wgmax = wgmax
        self.exCoeff = exCoeff
        self.maxdx = maxdx
        self.g = g
        self.cp = cp
        self.vk = vk
        self.r = r
        self.rv = rv
        self.lv = lv
        self.pi = pi
        self.sigma = sigma
        self.waterDens = waterDens
        self.lvtt = lvtt
        self.tt = tt
        self.estt = estt
        self.cl = cl
        self.cpv = cpv
        self.b = b
        self.cm = cm
        self.colburn = colburn

    def __repr__(self):
        return 'Parameter,\n dayBLHeight: {}\n nightBLHeight: {}\n refHeight: {}\n ' \
            'tempHeight: {}\n windHeight: {}\n circCoeff: {}\n dayThreshold: {}\n ' \
            'nightThreshold: {}\n treeFLat: {}\n grassFLat: {}\n vegAlbedo: {}\n ' \
            'vegStart: {}\n vegEnd: {}\n nightSetStart: {}\n nightSetEnd: {}\n ' \
            'windMin: {}\n wgmax: {}\n exCoeff: {}\n maxdx: {}\n'.format(
                self.dayBLHeight, self.nightBLHeight, self.refHeight, self.tempHeight,
                self.windHeight, self.circCoeff, self.dayThreshold, self.nightThreshold,
                self.treeFLat, self.grassFLat, self.vegAlbedo, self.vegStart,
                self.vegEnd, self.nightSetStart, self.nightSetEnd, self.windMin,
                self.wgmax, self.exCoeff, self.maxdx)



"""Functions for sychrometric calculations."""
from __future__ import division

from math import log, pow, exp


def psychrometrics(Tdb_in, w_in, P):
    """Modified version of Psychometrics by Tea Zakula MIT Building Technology Lab.

    Args:
        Tdb_in: Number for dry bulb temperature (K).
        w_in: Number for Humidity Ratio (kgv/kgda).
        P: Atmospheric Station Pressure (P).

    Returns:
        A tuple with six values

        - Tdb: Number for dry bulb temperature in (C).

        - w: Number for Humidity Ratio (kgv/kgda).

        - phi: Number for relative humidity (Pw/Pws*100).

        - h: Number for enthalpy (J/kga).

        - Tdp: Number for dew point temperature (C).

        - v: Number for specific volume (m3/kga).
    """
    # Change units
    c_air = 1006.  # [J/kg] air heat capacity, value from ASHRAE Fundamentals
    hlg = 2501000.  # [J/kg] latent heat, value from ASHRAE Fundamentals
    cw = 1860.  # [J/kg] value from ASHRAE Fundamentals
    P = P / 1000.  # convert from Pa to kPa

    Tdb = Tdb_in - 273.15
    w = w_in

    # phi (RH) calculation from Tdb and w
    Pw = (w * P) / (0.621945 + w)  # partial pressure of water vapor
    Pws = saturation_pressure(Tdb)  # Get saturation pressure for given Tdb
    phi = Pw / Pws * 100.0

    # enthalpy calculation from Tdb and w
    h = c_air * Tdb + w * (hlg + cw * Tdb)  # [J kga-1]

    # specific volume calculation from Tdb and w
    v = 0.287042 * (Tdb + 273.15) * (1 + 1.607858 * w) / P

    # dew point calculation from w
    _pw = (w * P) / (0.621945 + w)  # water vapor partial pressure in kPa
    alpha = log(_pw)

    Tdp = 6.54 + 14.526 * alpha + pow(alpha, 2) * 0.7389 + pow(alpha, 3) * 0.09486 + \
        pow(_pw, 0.1984) * 0.4569  # valid for Tdp between 0 C and 93 C

    return Tdb, w, phi, h, Tdp, v


def saturation_pressure(Tdb_):
    """Saturation pressure from dry bulb pressure.

    Args:
        Tdb_: Dry bulb temperature (K).

    Returns:
        Saturation pressure.
    """
    T = Tdb_ + 273.15

    # N.B In Matlab, negative values are converted to complex values.
    # log(-x) = log(x) + log(-1) = log(x) + i*pi
    # Python will throw an exception. Negative value occurs here if simulation
    # timestep (dtSim) is large, i.e 3600s.
    _Pws = exp(-1 * 5.8002206e3 / T + 1.3914993 + 4.8640239e-2 * T * -1.0 +
               4.1764768e-5 * pow(T, 2) - 1.4452093e-8 * pow(T, 3) +
               6.5459673 * log(T))  # Pa

    return _Pws / 1000.  # kPa


def moist_air_density(P, Tdb, H):
    """Moist air density (kgv m-3) from dry bulb temp, HR, and pressure[1][2][3].

    Note:
        [1] ASHRAE Fundamentals (2005) ch. 6 eqn. 28
        [2] ASHRAE Fundamentals (2009) ch. 1 eqn. 28
        [3] https://github.com/psychrometrics/Libraries/blob/master/Psychrometrics_SI.cpp

    Args:
        P: Number for Pressure (P).
        Tdb: Number for dry bulb temperature in (C).
        H: Number for Humidity Ratio (kgv/kgda).

    Returns:
        Moist air density (kgv m-3).
    """

    return P / (1000 * 0.287042 * Tdb * (1. + 1.607858 * H))


def hum_from_rhum_temp(RH, T, P):
    """Specific Humidity (kgh20/kgn202) from RH, T and Pa.

    Args:
        RH: Number for Relative Humidity.
        T: Number for dry bulb temperature (C).
        P: Number for air pressure (P).

    Returns:
        Specific Humidity (kgh20/kgn202).
    """
    # Saturation vapour pressure from ASHRAE
    C8 = -5.8002206e3
    C9 = 1.3914993
    C10 = -4.8640239e-2
    C11 = 4.1764768e-5
    C12 = -1.4452093e-8
    C13 = 6.5459673

    T += 273.15

    PWS = exp(C8 / T + C9 + C10 * T + C11 * pow(T, 2) + C12 * pow(T, 3) + C13 * log(T))
    PW = RH * PWS / 100.0  # Vapour pressure

    return 0.62198 * PW / (P - PW)  # 4. Specific humidity



"""Test SimParam class."""

import os
import pytest
from .test_base import auto_setup_uwg, setup_open_matlab_ref, TEST_DIR
from uwg import SimParam


def test_uwg_simparam_matlab_init():
    """Matlab value comparison for simparam initialization in main uwg."""

    # Initialize uwg from default initialize.uwg for 31 days simulation
    # Jan 1 00:00 - Feb 1 00:00

    # Month,1,        # starting month (1-12)
    # Day,1,          # starting day (1-31)
    # nDay,31,        # number of days to run simultion
    # dtSim,300,      # simulation time step (s)
    # dtWeather,3600, # weather time step (s)

    testuwg = auto_setup_uwg()
    testuwg.generate()

    # open matlab ref file
    uwg_matlab_val = \
        setup_open_matlab_ref('matlab_simparam', 'matlab_ref_simparam_init.txt')

    uwg_python_val = [
        testuwg.simTime.dt,            # uwg time simulation time step
        testuwg.simTime.timeForcing,   # weather data timestep
        testuwg.simTime.month,
        testuwg.simTime.day,
        testuwg.simTime.days,
        testuwg.simTime.timePrint,     # weather data timestep
        testuwg.simTime.timeDay,       # how many times weather senses in a day
        testuwg.simTime.timeSim,       # how many steps in weather data simulation
        testuwg.simTime.timeMax,       # total seconds in simulation days
        testuwg.simTime.nt,            # total number of timesteps for uwg simuation
        testuwg.simTime.julian,
        testuwg.simTime.timeInitial,   # sensor data in epw for intial time
        testuwg.simTime.timeFinal,     # sensor data in epw for final time
        testuwg.simTime.secDay,        # current seconds in day
        testuwg.simTime.hourDay]

    # matlab ref checking
    assert len(uwg_matlab_val) == len(uwg_python_val)
    for i in range(len(uwg_matlab_val)):
        assert uwg_python_val[i] == \
            pytest.approx(uwg_matlab_val[i], abs=1e-15), 'error at index={}'.format(i)


def test_uwg_simparam_matlab_update_date():
    """ Matlab value comparison for simparam UpdateDate function in main uwg """

    # Initialize uwg from initialize_simparam.uwg for 150 day simulation
    # Mar 15 00:00 - Aug 12th, 00:00

    # Month,3,        # starting month (1-12)
    # Day,15,         # starting day (1-31)
    # nDay,150,       # number of days to run simultion
    # dtSim,300,      # simulation time step (s)
    # dtWeather,3600, # weather time step (s)

    param_path = os.path.join(TEST_DIR, 'parameters', 'initialize_simparam.uwg')
    testuwg = auto_setup_uwg(param_path=param_path)

    testuwg.generate()
    testuwg.simulate()

    # open matlab ref file
    uwg_matlab_val = \
        setup_open_matlab_ref('matlab_simparam', 'matlab_ref_simparam_update_date.txt')

    uwg_python_val = [
        testuwg.simTime.secDay,
        testuwg.simTime.day,
        testuwg.simTime.julian,
        testuwg.simTime.month,
        testuwg.simTime.day,
        testuwg.simTime.hourDay,
        # Add 1 to keep consistent with matlab list convention
        testuwg.ceil_time_step + 1]

    # matlab ref checking
    assert len(uwg_matlab_val) == len(uwg_python_val)
    for i in range(len(uwg_matlab_val)):
        assert uwg_python_val[i] == \
            pytest.approx(uwg_matlab_val[i], abs=1e-15), 'error at index={}'.format(i)


def test_simparam():
    """Tests simparam.py."""

    dtSim = 300             # Sim time step
    dtWeather = 3600        # Weather data time-step
    MONTH = 7               # Begin month
    DAY = 30                # Begin day of the month
    NUM_DAYS = 7            # Number of days of simulation

    simTime = SimParam(dtSim, dtWeather, MONTH, DAY, NUM_DAYS)

    # Simulation Parameters tests
    assert simTime.timeSim == pytest.approx(168, abs=1e-6)
    assert simTime.timeMax == pytest.approx(604800, abs=1e-6)
    assert simTime.nt == pytest.approx(2017, abs=1e-6)

    # Test UpdateDate() for < 1 hr
    for i in range(11):  # 11 * 300 = 3300 seconds = 55min
        simTime.update_date()
    assert simTime.secDay == pytest.approx(3300., abs=1e-6)
    assert simTime.day == pytest.approx(30., abs=1e-6)
    assert simTime.hourDay == pytest.approx(0., abs=1e-6)

    # for == 1 hr
    simTime.update_date()
    assert simTime.secDay == pytest.approx(3600., abs=1e-6)
    assert simTime.hourDay == pytest.approx(1., abs=1e-6)

    # for > 24hr
    for i in range(23 * 12):
        simTime.update_date()
    assert simTime.secDay == pytest.approx(0., abs=1e-6)
    assert simTime.day == pytest.approx(31., abs=1e-6)
    assert simTime.hourDay == pytest.approx(0., abs=1e-6)

    # for == 1 month
    for i in range(24 * 12):
        simTime.update_date()
    assert simTime.secDay == pytest.approx(0., abs=1e-6)
    assert simTime.day == pytest.approx(1., abs=1e-6)
    assert simTime.hourDay == pytest.approx(0., abs=1e-6)
    assert simTime.month == pytest.approx(8, abs=1e-6)

    # for + 1 month
    for i in range(24 * 12 * 31):
        simTime.update_date()
    assert simTime.secDay == pytest.approx(0., abs=1e-6)
    assert simTime.day == pytest.approx(1., abs=1e-6)
    assert simTime.hourDay == pytest.approx(0., abs=1e-6)
    assert simTime.month == pytest.approx(9, abs=1e-6)


"""Class for Schedule Definition."""

from .utilities import float_positive, REF_BUILTERA_SET, REF_BUILTERA


try:
    str = basestring
except NameError:
    pass


class SchDef(object):
    """Schedule definition class.

    The internal load weekly schedules consists of three lists of 24 values representing hours in
    weekday, Saturday, and Sunday.

    Args:
        elec: Weekly schedule of fractional electricity plug process loads.
        gas: Weekly schedule of fractional gas process loads.
        light: Weekly schedule of fractional light process loads.
        occ: Weekly schedule of fractional occupant number.
        cool: Weekly schedule of cooling temperatures.
        heat: Weekly schedule of heating temperatures.
        q_elec: Maximum electrical plug process load [W/m2].
        q_gas: Maximum gas process load per unit area [W/m2].
        q_light: Maximum light process load per unit area [W/m2].
        n_occ: Maximum number of occupants per unit area [person/m2].
        vent: Maximum ventilation rate per unit area [m3/s/m2].
        swh: Optional property for weekly schedule of fractional hot water rate. This
            property will be a weekly schedule of zero values, as default.
        v_swh: Optional property for maximum volumetric hot water rate per unit area
            [L/hr/m2]. (Default: 0).
        bldtype: Text referring to a building type. By default, 16 building types are
            defined in the UWG according to models from the Department of Energy (DOE).
            Custom building types can also be defined with a new name. Note that this
            value along with the BEMDef builtera must exactly match the identifiers in
            the UWG bld list in order to specify the fraction of total built stock the
            building occupies in the UWG simulation. Choose from the following to
            reference or overwrite a BEM associated with a DOE reference building type:
            'fullservicerestaurant', 'hospital', 'largehotel', 'largeoffice',
            'mediumoffice', 'midriseapartment', 'outpatient', 'primaryschool',
            'quickservicerestaurant', 'secondaryschool', 'smallhotel', 'smalloffice',
            'standaloneretail', 'stripmall', 'supermarket', or 'warehouse'.
        builtera: Text defining building built era. Must be one of the following:
            "pre80" (pre-1980s), "pst80" (post-1980s), or "new" (new construction).
            This value along with the bldtype must exactly match the identifiers in
            the bld array in order to specify the fraction of total built stock the
            building occupies in the UWG simulation.

    Properties:
        * elec
        * gas
        * light
        * occ
        * cool
        * heat
        * swh
        * q_elec
        * q_gas
        * q_light
        * n_occ
        * vent
        * v_swh
        * bldtype
        * builtera
        * zonetype
    """
    DEFAULT_SWH = [[0 for j in range(24)] for i in range(3)]

    def __init__(self, elec, gas, light, occ, cool, heat, q_elec, q_gas, q_light,
                 n_occ, vent, bldtype, builtera, swh=DEFAULT_SWH, v_swh=0):
        self.elec = elec
        self.gas = gas
        self.light = light
        self.occ = occ
        self.cool = cool
        self.heat = heat
        self.q_elec = q_elec
        self.q_gas = q_gas
        self.q_light = q_light
        self.n_occ = n_occ
        self.vent = vent
        self.v_swh = v_swh
        self.swh = swh
        # Properties to be set in readDOE
        self.bldtype = bldtype  # DOE reference building type
        self.builtera = builtera  # pre80, pst80, new
        self.zonetype = None  # climate zone number (only used in testing).

    @property
    def elec(self):
        """Get or set weekly schedule of fractional electricity plug process loads.

        Weekly schedule consists of three lists of 24 values representing hours in
        weekday, Saturday, and Sunday.
        """
        return self._elec

    @elec.setter
    def elec(self, value):
        self._elec = SchDef.check_week_validity(value, 'elec')

    @property
    def gas(self):
        """Get or set weekly schedule of fractional gas process loads.

        Weekly schedule consists of three lists of 24 values representing hours in
        weekday, Saturday, and Sunday.
        """
        return self._gas

    @gas.setter
    def gas(self, value):
        self._gas = SchDef.check_week_validity(value, 'gas')

    @property
    def light(self):
        """Get or set weekly schedule of fractional light process loads.

        Weekly schedule consists of three lists of 24 values representing hours in
        weekday, Saturday, and Sunday.
        """
        return self._light

    @light.setter
    def light(self, value):
        self._light = SchDef.check_week_validity(value, 'light')

    @property
    def occ(self):
        """Get or set weekly schedule of occupant number.

        Weekly schedule consists of three lists of 24 values representing hours in
        weekday, Saturday, and Sunday.
        """
        return self._occ

    @occ.setter
    def occ(self, value):
        self._occ = SchDef.check_week_validity(value, 'occ')

    @property
    def cool(self):
        """Get or set weekly schedule of cooling temperatures.

        Weekly schedule consists of three lists of 24 values representing hours in
        weekday, Saturday, and Sunday.
        """
        return self._cool

    @cool.setter
    def cool(self, value):
        self._cool = SchDef.check_week_validity(value, 'cool')

    @property
    def heat(self):
        """Get or set weekly schedule of heating temperatures.

        Weekly schedule consists of three lists of 24 values representing hours in
        weekday, Saturday, and Sunday.
        """
        return self._heat

    @heat.setter
    def heat(self, value):
        self._heat = SchDef.check_week_validity(value, 'heat')

    @property
    def swh(self):
        """Get or set weekly schedule of fractional hot water rate.

        Weekly schedule consists of three lists of 24 values representing hours in
        weekday, Saturday, and Sunday.
        """
        return self._swh

    @swh.setter
    def swh(self, value):
        self._swh = SchDef.check_week_validity(value, 'swh')

    @property
    def q_elec(self):
        """Get or set maximum electrical plug process load [W/m2]."""
        return self._q_elec

    @q_elec.setter
    def q_elec(self, value):
        self._q_elec = float_positive(value, 'q_elec')

    @property
    def q_gas(self):
        """Get or set maximum gas process load per unit area [W/m2]."""
        return self._q_gas

    @q_gas.setter
    def q_gas(self, value):
        self._q_gas = float_positive(value, 'q_gas')

    @property
    def q_light(self):
        """Get or set maximum light process load per unit area [W/m2]."""
        return self._q_light

    @q_light.setter
    def q_light(self, value):
        self._q_light = float_positive(value, 'q_light')

    @property
    def n_occ(self):
        """Get or set maximum number of occupants per unit area [person/m2]."""
        return self._n_occ

    @n_occ.setter
    def n_occ(self, value):
        self._n_occ = float_positive(value, 'n_occ')

    @property
    def vent(self):
        """Get or set maximum ventilation rate per unit area [m3/s/m2]."""
        return self._vent

    @vent.setter
    def vent(self, value):
        self._vent = float_positive(value, 'vent')

    @property
    def v_swh(self):
        """Get or set maximum volumetric hot water rate per unit area [L/hr/m2]."""
        return self._v_swh

    @v_swh.setter
    def v_swh(self, value):
        self._v_swh = float_positive(value, 'v_swh')

    @property
    def bldtype(self):
        """Get or set text for building type.

        By default, 16 building types are defined in the UWG according to models from
        the Department of Energy (DOE). Choose from the following to reference or
        overwrite a schedule associated with a DOE reference building type:

        * 'fullservicerestaurant'
        * 'hospital'
        * 'largehotel'
        * 'largeoffice'
        * 'medoffice'
        * 'midriseapartment'
        * 'outpatient'
        * 'primaryschool'
        * 'quickservicerestaurant'
        * 'secondaryschool'
        * 'smallhotel'
        * 'smalloffice'
        * 'standaloneretail'
        * 'stripmall'
        * 'supermarket'
        * 'warehouse'

        Custom building types can also be defined with a new name. If a custom SchDef is
        defined with the same name as a reference DOE building type from the list above,
        the reference SchDef will be overwritten by the custom SchDef. Note that this
        value along with the SchDef builtera must exactly match the identifiers in the
        UWG bld list in order to specify the fraction of total built stock the building
        occupies in the UWG simulation.
        """
        return self._bldtype

    @bldtype.setter
    def bldtype(self, value):
        assert isinstance(value, str), 'The bldtype must be a string. ' \
            'Got: {}.'.format(value.lower())
        self._bldtype = value

    @property
    def builtera(self):
        """Get or set text for built era.

        Must be one of the following:

        * 'pre80' - pre-1980s
        * 'pst80' - post-1980s
        * 'new' - new construction

        This value along with the bldtype must exactly match the identifiers in
        the bld array in order to specify the fraction of total built stock the
        building occupies in the UWG simulation.
        """
        return self._builtera

    @builtera.setter
    def builtera(self, value):
        assert isinstance(value, str) and value in REF_BUILTERA_SET, \
            'The builtera must be one of {}.Got: {}.'.format(
                REF_BUILTERA, value.lower())
        self._builtera = value

    @classmethod
    def from_dict(cls, data):
        """Create a SchDef object from a dictionary.

        Args:
            data: A SchDef dictionary following the format below.

        .. code-block:: python

            _example_week = [
                [0.2, 0.2, 0.2, 0.2, 0.2, 0.4, 0.7, 0.9, 0.9, 0.6, 0.6, 0.6, 0.6, 0.6,
                 0.7, 0.8, 0.9, 0.9, 0.8, 0.8, 0.7, 0.3, 0.2, 0.2],  # Weekday
                [0.2, 0.2, 0.2, 0.2, 0.2, 0.3, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5,
                 0.6, 0.7, 0.7, 0.7, 0.7, 0.5, 0.4, 0.3, 0.2, 0.2],  # Saturday
                [0.2, 0.2, 0.2, 0.2, 0.2, 0.3, 0.4, 0.4, 0.4, 0.4, 0.4, 0.4, 0.4, 0.4,
                 0.4, 0.4, 0.4, 0.4, 0.4, 0.4, 0.3, 0.3, 0.2, 0.2]]  # Sunday

            {
            "elec": _example_week,
            "gas": _example_week,
            "light": _example_week,
            "occ": _example_week,
            "cool": _example_week,
            "heat": _example_week,
            "swh": _example_week,
            "q_elec" = q_elec,
            "q_gas" = q_gas,
            "q_light" = q_light,
            "n_occ" = n_occ,
            "vent" = vent,
            "v_swh" = v_swh,
            "bldtype": 0,
            "builtera": 1
            }
        """
        assert data['type'] == 'SchDef', 'Expected ' \
            'SchDef dictionary. Got {}.'.format(data['type'])

        return cls(elec=data['elec'], gas=data['gas'], light=data['light'],
                   occ=data['occ'], cool=data['cool'], heat=data['heat'],
                   swh=data['swh'], q_elec=data['q_elec'], q_gas=data['q_gas'],
                   q_light=data['q_light'], n_occ=data['n_occ'], vent=data['vent'],
                   v_swh=data['v_swh'], bldtype=data['bldtype'],
                   builtera=data['builtera'])

    def to_dict(self):
        """SchDef dictionary representation."""
        base = {'type': 'SchDef'}
        base['elec'] = self.elec
        base['gas'] = self.gas
        base['light'] = self.light
        base['occ'] = self.occ
        base['cool'] = self.cool
        base['heat'] = self.heat
        base['swh'] = self.swh
        base['q_elec'] = self.q_elec
        base['q_gas'] = self.q_gas
        base['q_light'] = self.q_light
        base['n_occ'] = self.n_occ
        base['vent'] = self.vent
        base['v_swh'] = self.v_swh
        base['bldtype'] = self.bldtype
        base['builtera'] = self.builtera
        return base

    @staticmethod
    def check_week_validity(week, name):
        assert isinstance(week, (list, tuple)), 'The {} property must be a ' \
            'list or tuple. Got {}.'.format(name, week)

        assert len(week) == 3, 'The {} property must be a 3 x 24 matrix. Got ' \
            '{} rows.'.format(name, len(week))

        for i, day in enumerate(week):
            assert len(day) == 24, 'The {} property must be a 3 x 24 ' \
                'matrix. Got {} columns for row {}.'.format(name, len(day), i)
            for val in day:
                assert isinstance(val, (float, int)), 'The {} property ' \
                    'must contain 3 lists of numbers. Got : {}.'.format(
                        name, val)
        return week

    def __repr__(self):
        return 'Schedule, bldtype: {}\n builtera: {}\n q_elec: {}\n q_gas: {}\n ' \
            'q_light: {}\n n_occ: {}\n vent: {}\n v_swh: {}\n elec: {}\n ' \
            'gas: {}\n light: {}\n occ: {}\n cool: {}\n heat: {}\n swh: {}\n'.format(
                self.bldtype, self.builtera, self.q_elec, self.q_gas, self.q_light,
                self.n_occ, self.vent, self.v_swh, self.elec, self.gas, self.light,
                self.occ, self.cool, self.heat, self.swh)



"""Class for Weather data from EPW."""
from .utilities import read_csv, str2fl
from .psychrometrics import hum_from_rhum_temp

try:
    range = xrange
except NameError:
    pass


class Weather(object):
    """Read EPW file[1] and store weather timeseries data.

    Args:
        climate_file: Text string for the name of the rural epw file that will be
            morphed.
        HI: Number for initial EPW row based on final Julian data.
        HF: Number for final EPW row based on start Julian data.

    Properties:
        * location -- location row from EPW file.
        * staTemp --  drybulb (K)
        * staTdp -- dewpoint (C)
        * staRhum -- air RH (%)
        * staPres -- air pressure (Pa)
        * staInfra -- horizontal Infrared Radiation Intensity (W m-2)
        * staHor -- horizontal radiation [W m-2]
        * staDir -- normal solar direct radiation (W m-2)
        * staDif -- horizontal solar diffuse radiation (W m-2)
        * staUdir -- wind direction
        * staUmod -- wind speed (m s-1)
        * staRobs -- precipitation (mm h-1)
        * staHum -- specific humidty (kgH20 kgN202-1)

    Note:
        [1] EPW CSV Format (In/Out). Retrieved August 26, 2020,
        from http://bigladdersoftware.com/epx/docs/8-2/auxiliary-programs/epw-csv-format-inout.html
    """

    def __init__(self, epw_path, HI, HF):
        try:
            self._climate_data = read_csv(epw_path)
        except Exception as e:
            raise Exception("Failed to read .epw file! {}".format(e.message))

        self.location = self._climate_data[0][1]
        cd = self._climate_data[HI:HF + 1]
        self.staTemp = str2fl([cd[i][6] for i in range(len(cd))])
        self.staTdp = str2fl([cd[i][7] for i in range(len(cd))])
        self.staRhum = str2fl([cd[i][8] for i in range(len(cd))])
        self.staPres = str2fl([cd[i][9] for i in range(len(cd))])
        # staInfra:
        self.staInfra = str2fl([cd[i][12] for i in range(len(cd))])
        # staHor:
        self.staHor = str2fl([cd[i][13] for i in range(len(cd))])
        # staDir:
        self.staDir = str2fl([cd[i][14] for i in range(len(cd))])
        # staDif:
        self.staDif = str2fl([cd[i][15] for i in range(len(cd))])
        self.staUdir = str2fl([cd[i][20] for i in range(len(cd))])
        self.staUmod = str2fl([cd[i][21] for i in range(len(cd))])
        self.staHum = [0.0] * len(self.staTemp)
        for i in range(len(self.staTemp)):
            self.staHum[i] = \
                hum_from_rhum_temp(self.staRhum[i], self.staTemp[i], self.staPres[i])
        self.staTemp = [s + 273.15 for s in self.staTemp]  # air temperature (K)

        # Set precipitation to array of zeros. This is done to avoid errors
        # for EPW files with incomplete or missing precipitation data. The
        # current implementation of the UWG doesn't use precipitation data
        # due to the difficulty in validating latent heat impact, the poor
        # quality precipitation data from weather sensors, and the marginal
        # impact it has on UHI.
        self.staRobs = [0.0] * len(self.staTemp)

    def __repr__(self):
        return 'Weather,\n City: {}\n, Max Tdb: {}C\n, Min Tdb: {}C\n'.format(
            self.location, round(max(self.staTemp) - 273.15, 2),
            round(min(self.staTemp) - 273.15, 2))



from __future__ import division, print_function

try:
    range = xrange
except NameError:
    pass

import sys
import os

try:
    import cPickle as pickle
except ImportError:
    import pickle

from .building import Building
from .material import Material
from .element import Element
from .BEMDef import BEMDef
from .schdef import SchDef
from .utilities import read_csv, str2fl
# to avoid circular imports load from util
from .utilities import REF_BLDTYPE, REF_BUILTERA, REF_ZONETYPE


DIR_CURR = os.path.abspath(os.path.dirname(__file__))
DIR_DOE_PATH = os.path.join(DIR_CURR, "..", "resources", "DOERefBuildings")

# Define standards: 16 buiding types, 3 built eras, 16 climate zones


def readDOE(serialize_output=True):
    """
    Read csv files of DOE buildings
    Sheet 1 = BuildingSummary
    Sheet 2 = ZoneSummary
    Sheet 3 = LocationSummary
    Sheet 4 = Schedules
    Note BLD8 & 10 = school


    Then make matrix of ref data as nested nested lists [16, 3, 16]:
    matrix refDOE = Building objs
    matrix Schedule = SchDef objs
    matrix refBEM (16,3,16) = BEMDef
    where:
        [16,3,16] is Type = 1-16, Era = 1-3, climate zone = 1-16
        i.e.
        Type: FullServiceRestaurant, Era: Pre80, Zone: 6A Minneapolis
    Nested tree:
    [TYPE_1:
        ERA_1:
            CLIMATE_ZONE_1
            ...
            CLIMATE_ZONE_16
        ERA_2:
            CLIMATE_ZONE_1
            ...
            CLIMATE_ZONE_16
        ...
        ERA_3:
            CLIMATE_ZONE_1
            ...
            CLIMATE_ZONE_16]

    """

    # 16 x 3 x 16 matrix of Building, SchDef, BEMDef objects
    refDOE = [[[None] * 16 for k_ in range(3)]
              for j_ in range(16)]  # refDOE: Building
    # Schedule: SchDef
    Schedule = [[[None] * 16 for k_ in range(3)] for j_ in range(16)]
    refBEM = [[[None] * 16 for k_ in range(3)]
              for j_ in range(16)]  # refBEM: BEMDef

    # Purpose: Loop through every DOE reference csv and extract building data
    # Nested loop = 16 types, 3 era, 16 zones = time complexity O(n*m*k) = 768
    for i in range(16):

        # i = 16 types of buildings
        # print "\tType: {} @i={}".format(BLDTYPE[i], i)

        # Read building summary (Sheet 1)
        file_doe_name_bld = os.path.join("{}".format(DIR_DOE_PATH), "BLD{}".format(
            i+1), "BLD{}_BuildingSummary.csv".format(i+1))
        list_doe1 = read_csv(file_doe_name_bld)
        # listof(listof 3 era values)
        # Number of Floors, this will be list of floats and str if "basement"
        nFloor = str2fl(list_doe1[3][3:6])
        glazing = str2fl(list_doe1[4][3:6])      # [?] Total
        hCeiling = str2fl(list_doe1[5][3:6])      # [m] Ceiling height
        ver2hor = str2fl(list_doe1[7][3:6])      # Wall to Skin Ratio
        # [m2] Gross Dimensions - Total area
        AreaRoof = str2fl(list_doe1[8][3:6])

        # Read zone summary (Sheet 2)
        file_doe_name_zone = os.path.join("{}".format(
            DIR_DOE_PATH), "BLD{}".format(i+1), "BLD{}_ZoneSummary.csv".format(i+1))
        list_doe2 = read_csv(file_doe_name_zone)
        # listof(listof 3 eras)
        AreaFloor = str2fl([list_doe2[2][5], list_doe2[3]
                            [5], list_doe2[4][5]])       # [m2]
        Volume = str2fl([list_doe2[2][6], list_doe2[3][6],
                         list_doe2[4][6]])       # [m3]
        AreaWall = str2fl([list_doe2[2][8], list_doe2[3][8],
                           list_doe2[4][8]])       # [m2]
        AreaWindow = str2fl([list_doe2[2][9], list_doe2[3]
                             [9], list_doe2[4][9]])       # [m2]
        Occupant = str2fl([list_doe2[2][11], list_doe2[3][11],
                           list_doe2[4][11]])    # Number of People
        Light = str2fl([list_doe2[2][12], list_doe2[3]
                        [12], list_doe2[4][12]])    # [W/m2]
        # [W/m2] Electric Plug and Process
        Elec = str2fl([list_doe2[2][13], list_doe2[3][13], list_doe2[4][13]])
        # [W/m2] Gas Plug and Process
        Gas = str2fl([list_doe2[2][14], list_doe2[3][14], list_doe2[4][14]])
        # [Litres/hr] Peak Service Hot Water
        SHW = str2fl([list_doe2[2][15], list_doe2[3][15], list_doe2[4][15]])
        # [L/s/m2] Ventilation
        Vent = str2fl([list_doe2[2][17], list_doe2[3][17], list_doe2[4][17]])
        # Air Changes Per Hour (ACH) Infiltration
        Infil = str2fl([list_doe2[2][20], list_doe2[3][20], list_doe2[4][20]])

        # Read location summary (Sheet 3)
        file_doe_name_location = os.path.join("{}".format(
            DIR_DOE_PATH), "BLD{}".format(i+1), "BLD{}_LocationSummary.csv".format(i+1))
        list_doe3 = read_csv(file_doe_name_location)
        # (listof (listof 3 eras (listof 16 climate types)))
        TypeWall = [list_doe3[3][4:20], list_doe3[14][4:20],
                    list_doe3[25][4:20]]            # Construction type
        RvalWall = str2fl([list_doe3[4][4:20], list_doe3[15]
                           [4:20], list_doe3[26][4:20]])     # [m2*K/W] R-value
        TypeRoof = [list_doe3[5][4:20], list_doe3[16][4:20],
                    list_doe3[27][4:20]]            # Construction type
        RvalRoof = str2fl([list_doe3[6][4:20], list_doe3[17]
                           [4:20], list_doe3[28][4:20]])     # [m2*K/W] R-value
        Uwindow = str2fl([list_doe3[7][4:20], list_doe3[18]
                          [4:20], list_doe3[29][4:20]])     # [W/m2*K] U-factor
        SHGC = str2fl([list_doe3[8][4:20], list_doe3[19][4:20],
                       list_doe3[30][4:20]])     # [-] coefficient
        HVAC = str2fl([list_doe3[9][4:20], list_doe3[20][4:20],
                       list_doe3[31][4:20]])     # [kW] Air Conditioning
        HEAT = str2fl([list_doe3[10][4:20], list_doe3[21][4:20],
                       list_doe3[32][4:20]])    # [kW] Heating
        COP = str2fl([list_doe3[11][4:20], list_doe3[22][4:20],
                      list_doe3[33][4:20]])    # [-] Air Conditioning COP
        # [%] Heating Efficiency
        EffHeat = str2fl([list_doe3[12][4:20], list_doe3[23]
                          [4:20], list_doe3[34][4:20]])
        # [m3/s] Fan Max Flow Rate
        FanFlow = str2fl([list_doe3[13][4:20], list_doe3[24]
                          [4:20], list_doe3[35][4:20]])

        # Read Schedules (Sheet 4)
        file_doe_name_schedules = os.path.join("{}".format(
            DIR_DOE_PATH), "BLD{}".format(i+1), "BLD{}_Schedules.csv".format(i+1))
        list_doe4 = read_csv(file_doe_name_schedules)

        # listof(listof weekday, sat, sun (list of 24 fractions)))
        # Equipment Schedule 24 hrs
        SchEquip = str2fl([list_doe4[1][6:30], list_doe4[2]
                           [6:30], list_doe4[3][6:30]])
        # Light Schedule 24 hrs; Wkday=Sat=Sun=Hol
        SchLight = str2fl([list_doe4[4][6:30], list_doe4[5]
                           [6:30], list_doe4[6][6:30]])
        # Occupancy Schedule 24 hrs
        SchOcc = str2fl([list_doe4[7][6:30], list_doe4[8]
                         [6:30], list_doe4[9][6:30]])
        # Cooling Setpoint Schedule 24 hrs
        SetCool = str2fl([list_doe4[10][6:30], list_doe4[11]
                          [6:30], list_doe4[12][6:30]])
        # Heating Setpoint Schedule 24 hrs; summer design
        SetHeat = str2fl([list_doe4[13][6:30], list_doe4[14]
                          [6:30], list_doe4[15][6:30]])
        # Gas Equipment Schedule 24 hrs; wkday=sat
        SchGas = str2fl([list_doe4[16][6:30], list_doe4[17]
                         [6:30], list_doe4[18][6:30]])
        # Solar Water Heating Schedule 24 hrs; wkday=summerdesign, sat=winterdesgin
        SchSWH = str2fl([list_doe4[19][6:30], list_doe4[20]
                         [6:30], list_doe4[21][6:30]])

        for j in range(3):

            # j = 3 built eras
            # print"\tEra: {} @j={}".format(BUILTERA[j], j)

            for k in range(16):

                # k = 16 climate zones
                # print "\tClimate zone: {} @k={}".format(ZONETYPE[k], k)

                B = Building(
                    hCeiling[j],                        # floorHeight by era
                    1,                                  # intHeatNight
                    1,                                  # intHeatDay
                    0.1,                                # intHeatFRad
                    0.1,                                # intHeatFLat
                    Infil[j],                           # infil (ACH) by era
                    # vent (m^3/s/m^2) by era, converted from liters
                    Vent[j]/1000.,
                    glazing[j],                         # glazing ratio by era
                    # uValue by era, by climate type
                    Uwindow[j][k],
                    # SHGC, by era, by climate type
                    SHGC[j][k],
                    'AIR',                              # cooling condensation system type: AIR, WATER
                    # cop by era, climate type
                    COP[j][k],
                    # coolCap converted to W/m2 by era, climate type
                    (HVAC[j][k]*1000.0)/AreaFloor[j],
                    # heatEff by era, climate type
                    EffHeat[j][k],
                    293)                                # initialTemp at 20 C

                # Not defined in the constructor
                # heating Capacity converted to W/m2 by era, climate type
                B.heat_cap = (HEAT[j][k] * 1000.0) / AreaFloor[j]
                refDOE[i][j][k] = B

                # Define wall, mass(floor), roof
                # Reference from E+ for conductivity, thickness (reference below)

                # Material: (thermalCond, volHeat = specific heat * density)
                Concrete = Material(1.311, 836.8 * 2240, "Concrete")
                Insulation = Material(0.049, 836.8 * 265.0, "Insulation")
                Gypsum = Material(0.16, 830.0 * 784.9, "Gypsum")
                Wood = Material(0.11, 1210.0 * 544.62, "Wood")
                Stucco = Material(0.6918,  837.0 * 1858.0, "Stucco")

                # Wall (1 in stucco, concrete, insulation, gypsum)
                # Check TypWall by era, by climate
                if TypeWall[j][k] == "MassWall":
                    # Construct wall based on R value of Wall from refDOE and properties defined above
                    # 1" stucco, 8" concrete, tbd insulation, 1/2" gypsum
                    Rbase = 0.271087  # R val based on stucco, concrete, gypsum
                    Rins = RvalWall[j][k] - Rbase  # find insulation value
                    D_ins = Rins * Insulation.thermalcond  # depth of ins from m2*K/W * W/m*K = m
                    if D_ins > 0.01:
                        thickness = [0.0254, 0.0508, 0.0508,
                                     0.0508, 0.0508, D_ins, 0.0127]
                        layers = [Stucco, Concrete, Concrete,
                                  Concrete, Concrete, Insulation, Gypsum]
                    else:
                        # if it's less then 1 cm don't include in layers
                        thickness = [0.0254, 0.0508,
                                     0.0508, 0.0508, 0.0508, 0.0127]
                        layers = [Stucco, Concrete, Concrete,
                                  Concrete, Concrete, Gypsum]

                    wall = Element(0.08, 0.92, thickness, layers,
                                   0., 293., 0., "MassWall")

                    # If mass wall, assume mass floor (4" concrete)
                    # Mass (assume 4" concrete);
                    alb = 0.2
                    emis = 0.9
                    thickness = [0.054, 0.054]
                    concrete = Material(1.31, 2240.0*836.8, 'Concrete')
                    mass = Element(alb, emis, thickness, [
                                   concrete, concrete], 0, 293, 1, "MassFloor")

                elif TypeWall[j][k] == "WoodFrame":
                    # 0.01m wood siding, tbd insulation, 1/2" gypsum
                    Rbase = 0.170284091    # based on wood siding, gypsum
                    Rins = RvalWall[j][k] - Rbase
                    D_ins = Rins * Insulation.thermalcond  # depth of insulatino

                    if D_ins > 0.01:
                        thickness = [0.01, D_ins, 0.0127]
                        layers = [Wood, Insulation, Gypsum]
                    else:
                        thickness = [0.01, 0.0127]
                        layers = [Wood, Gypsum]

                    wall = Element(0.22, 0.92, thickness, layers,
                                   0., 293., 0., "WoodFrameWall")

                    # If wood frame wall, assume wooden floor
                    alb = 0.2
                    emis = 0.9
                    thickness = [0.05, 0.05]
                    wood = Material(1.31, 2240.0*836.8, 'Wood')
                    mass = Element(alb, emis, thickness, [
                                   wood, wood], 0., 293., 1., "WoodFloor")

                elif TypeWall[j][k] == "SteelFrame":
                    # 1" stucco, 8" concrete, tbd insulation, 1/2" gypsum
                    Rbase = 0.271087  # based on stucco, concrete, gypsum
                    Rins = RvalWall[j][k] - Rbase
                    D_ins = Rins * Insulation.thermalcond
                    if D_ins > 0.01:
                        thickness = [0.0254, 0.0508, 0.0508,
                                     0.0508, 0.0508, D_ins, 0.0127]
                        layers = [Stucco, Concrete, Concrete,
                                  Concrete, Concrete, Insulation, Gypsum]
                    else:    # If insulation is too thin, assume no insulation
                        thickness = [0.0254, 0.0508,
                                     0.0508, 0.0508, 0.0508, 0.0127]
                        layers = [Stucco, Concrete, Concrete,
                                  Concrete, Concrete, Gypsum]
                    wall = Element(0.15, 0.92, thickness, layers,
                                   0., 293., 0., "SteelFrame")

                    # If mass wall, assume mass foor
                    # Mass (assume 4" concrete),
                    alb = 0.2
                    emis = 0.93
                    thickness = [0.05, 0.05]
                    mass = Element(alb, emis, thickness, [
                                   Concrete, Concrete], 0., 293., 1., "MassFloor")

                elif TypeWall[j][k] == "MetalWall":
                    # metal siding, insulation, 1/2" gypsum
                    alb = 0.2
                    emis = 0.9
                    # use derived insul thickness or 0.01 based on max
                    D_ins = max(
                        (RvalWall[j][k] * Insulation.thermalcond) / 2, 0.01)
                    thickness = [D_ins, D_ins, 0.0127]
                    materials = [Insulation, Insulation, Gypsum]
                    wall = Element(alb, emis, thickness,
                                   materials, 0, 293, 0, "MetalWall")

                    # Mass (assume 4" concrete);
                    alb = 0.2
                    emis = 0.9
                    thickness = [0.05, 0.05]
                    concrete = Material(1.31, 2240.0*836.8, 'Concrete')
                    mass = Element(alb, emis, thickness, [
                                   concrete, concrete], 0., 293., 1., "MassFloor")

                # Roof
                if TypeRoof[j][k] == "IEAD":  # Insulation Entirely Above Deck
                    # IEAD-> membrane, insulation, decking
                    alb = 0.2
                    emis = 0.93
                    D_ins = max(RvalRoof[j][k] *
                                Insulation.thermalcond / 2., 0.01)
                    roof = Element(alb, emis, [D_ins, D_ins], [
                                   Insulation, Insulation], 0., 293., 1, "IEAD")

                elif TypeRoof[j][k] == "Attic":
                    # IEAD-> membrane, insulation, decking
                    alb = 0.2
                    emis = 0.9
                    D_ins = max(RvalRoof[j][k] *
                                Insulation.thermalcond / 2., 0.01)
                    roof = Element(alb, emis, [D_ins, D_ins], [
                                   Insulation, Insulation], 0., 293., 1, "Attic")

                elif TypeRoof[j][k] == "MetalRoof":
                    # IEAD-> membrane, insulation, decking
                    alb = 0.2
                    emis = 0.9
                    D_ins = max(RvalRoof[j][k] *
                                Insulation.thermalcond / 2., 0.01)
                    roof = Element(alb, emis, [D_ins, D_ins], [
                                   Insulation, Insulation], 0., 293., 1, "MetalRoof")

                # Define bulding energy model, set fraction of the urban floor space of this typology to zero
                bldtype, builtera, zonetype = \
                    REF_BLDTYPE[i], REF_BUILTERA[j], REF_ZONETYPE[k]
                refBEM[i][j][k] = BEMDef(building=B, mass=mass, wall=wall, roof=roof,
                                         bldtype=bldtype, builtera=builtera)
                refBEM[i][j][k].zonetype = zonetype  # for testing
                # max fan flow rate (m^3/s) per DOE
                refBEM[i][j][k].building.FanMax = FanFlow[j][k]

                # Calculate internal heat loads for schedule
                # W/m^2 (max) for electrical plug process
                q_elec = Elec[j]
                q_light = Light[j]                 # W/m^2 (max) for light
                n_occ = Occupant[j]/AreaFloor[j]   # Person/m^2
                q_gas = Gas[j]                     # W/m^2 (max) for gas
                vent = Vent[j]/1000.0             # m^3/s/m^2
                v_swh = SHW[j]/AreaFloor[j]        # L/hr/m2

                Schedule[i][j][k] = SchDef(elec=SchEquip, gas=SchGas, light=SchLight, occ=SchOcc, cool=SetCool,
                                           heat=SetHeat, swh=SchSWH, q_elec=q_elec, q_gas=q_gas, q_light=q_light,
                                           n_occ=n_occ, vent=vent, v_swh=v_swh, bldtype=bldtype, builtera=builtera)
                Schedule[i][j][k].zonetype = zonetype  # for testing only.

    # if not test serialize refDOE,refBEM,Schedule and store in resources
    if serialize_output:

        # create a binary file for serialized obj
        pkl_file_path = os.path.join(DIR_CURR, 'refdata', 'readDOE.pkl')
        pickle_readDOE = open(pkl_file_path, 'wb')

        # dump in ../resources
        # Pickle objects, protocol 1 b/c binary file
        # pickle.dump(refDOE, pickle_readDOE,1)  # Not used. Commented out.
        pickle.dump(refBEM, pickle_readDOE, 1)
        pickle.dump(Schedule, pickle_readDOE, 1)

        pickle_readDOE.close()

    return refBEM, Schedule


if __name__ == "__main__":

    # Set to True only if you want create new .pkls of DOE refs
    # Use --serialize switch to serialize the readDOE data
    if len(sys.argv) > 1 and sys.argv[1] == "--serialize":
        refBEM, Schedule = readDOE(True)
    else:
        refBEM, Schedule = readDOE(False)


# Material ref from E+
#     1/2IN Gypsum,            !- Name
#     Smooth,                  !- Roughness
#     0.0127,                  !- Thickness {m}
#     0.1600,                  !- Conductivity {W/m-K}
#     784.9000,                !- Density {kg/m3}
#     830.0000,                !- Specific Heat {J/kg-K}
#     0.9000,                  !- Thermal Absorptance
#     0.9200,                  !- Solar Absorptance
#     0.9200;                  !- Visible Absorptance
#
# Material,
#     1IN Stucco,              !- Name
#     Smooth,                  !- Roughness
#     0.0253,                  !- Thickness
#     0.6918,                  !- Conductivity
#     1858.0000,               !- Density
#     837.0000,                !- Specific Heat
#     0.9000,                  !- Thermal Absorptance
#     0.9200,                  !- Solar Absorptance
#     0.9200;                  !- Visible Absorptance
#
# Material,
#     8IN CONCRETE HW,  !- Name
#     Rough,                   !- Roughness
#     0.2032,                  !- Thickness {m}
#     1.3110,                  !- Conductivity {W/m-K}
#     2240.0000,               !- Density {kg/m3}
#     836.8000,                !- Specific Heat {J/kg-K}
#     0.9000,                  !- Thermal Absorptance
#     0.7000,                  !- Solar Absorptance
#     0.7000;                  !- Visible Absorptance
#
# Material,
#     Mass NonRes Wall Insulation, !- Name
#     MediumRough,             !- Roughness
#     0.0484268844343858,      !- Thickness {m}
#     0.049,                   !- Conductivity {W/m-K}
#     265.0000,                !- Density {kg/m3}
#     836.8000,                !- Specific Heat {J/kg-K}
#     0.9000,                  !- Thermal Absorptance
#     0.7000,                  !- Solar Absorptance
#     0.7000;                  !- Visible Absorptance
#
# Material,
#     Std Wood 6inch,          !- Name
#     MediumSmooth,            !- Roughness
#     0.15,                    !- Thickness {m}
#     0.12,                    !- Conductivity {W/m-K}
#     540.0000,                !- Density {kg/m3}
#     1210,                    !- Specific Heat {J/kg-K}
#     0.9000000,               !- Thermal Absorptance
#     0.7000000,               !- Solar Absorptance
#     0.7000000;               !- Visible Absorptance! Common Materials
#
# Material,
#     Wood Siding,             !- Name
#     MediumSmooth,            !- Roughness
#     0.0100,                  !- Thickness {m}
#     0.1100,                  !- Conductivity {W/m-K}
#     544.6200,                !- Density {kg/m3}
#     1210.0000,               !- Specific Heat {J/kg-K}
#     0.9000,                  !- Thermal Absorptance
#     0.7800,                  !- Solar Absorptance
#     0.7800;                  !- Visible Absorptance



"""Class for simulation time parameters."""
from __future__ import division, print_function

try:
    range = xrange
except NameError:
    pass

import math
from .utilities import is_near_zero


class SimParam(object):
    """Calculates simulation time parameters.

    Simulation time parameters are calculated based on initial date and weather
    data time step, and simulation timestep.

    Args:
        dt: Number for simulation time step in seconds.
        timefor: Number for weather data time-step in seconds.
        M: Number between 1 and 12 indicating the simulation start month.
        DAY: Number between 1 and 31 indicating the simulation start day.
        days: Number indicating number of days to simulate.

    Properties:
        * dt -- uwg time simulation time step
        * timeForcing -- weather data timestep
        * month -- start month
        * day -- start day
        * days -- number of days in simulation
        * timePrint -- weather data timestep
        * timeDay -- number of times weather senses in a day
        * timeSim -- number of steps in weather data simulation
        * timeMax -- total seconds in simulation days
        * nt -- total number of timesteps
        * inobis -- list of julian dates for the start of the months
        * julian -- simulation start julian date
        * timeInitial -- epw sensor data for initial time based on julian day & timesteps
        * timeFinal -- epw sensor data for final time based on julian day & timesteps
        * secDay -- current seconds in day
        * hourDay -- current hours in day
    """
    TIMESTEP_CONFLICT_MSG = "TIMESTEP ERROR! Timestep must be a factor of 3600."

    def __init__(self, dt, timefor, M, DAY, days):
        self.dt = dt
        self.timeForcing = timefor
        self.month = int(M)
        self.day = DAY
        self.days = days
        self.timePrint = timefor
        self.timeDay = 24 * 3600 / timefor
        self.timeSim = self.timeDay * days
        self.timeMax = 24. * 3600. * days
        self.nt = int(round(self.timeMax / self.dt + 1))
        self.inobis = [0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334]
        self.julian = self.inobis[self.month - 1] + DAY - 1
        # H1: (julian day * number of timesteps in a day) == sensor data index in epw
        H1 = int((self.inobis[self.month - 1] + DAY - 1) * self.timeDay)
        self.timeInitial = H1 + 8
        self.timeFinal = int(H1 + self.timeDay * self.days - 1 + 8)
        self.secDay = 0
        self.hourDay = 0

    def update_date(self):
        self.secDay = self.secDay + self.dt

        if is_near_zero(self.secDay - (3600 * 24)):
            self.day += 1
            self.julian = self.julian + 1
            self.secDay = 0.
            for j in range(12):
                if is_near_zero(self.julian - self.inobis[j]):
                    self.month = self.month + 1
                    self.day = 1

        if self.secDay > (3600 * 24):
            raise Exception("{}. CURRENTLY AT {}.".format(
                self.TIMESTEP_CONFLICT_MSG, self.dt))

        self.hourDay = int(math.floor(self.secDay / 3600.))  # 0 - 23hr

    def __repr__(self):
        return "SimParam: Start = {}/{}, days = {}, timestep = {}s".format(
            self.month, int(self.day), self.days, self.dt)



"""Class definition for the Urban Boundary Layer (UBL)."""
from __future__ import division

try:
    range = xrange
except NameError:
    pass

import logging
from .utilities import is_near_zero


class UBLDef(object):
    """Urban Boundary Layer (UBL) calculations.

    Args:
        location: Text string for relative location within a city. Choose from "N", "NE",
            "E", "SE", "S", "SW", "W", "NW", or "C".
        charLength: Value for characteristic length of the urban area [m].
        initialTemp: Value for initial temperature [K].
        maxdx: Value for maximum discretization length for the UBL model [m].
        dayBLHeight: Value for daytime mixing height [m].
        nightBLHeight: Value for nighttime boundary-layer height [m].

    Properties:
        * location -- relative location within a city (N, NE, E, SE, S, SW, W, NW, C).
        * charLength -- characteristic length of the urban area (m)
        * perimeter -- horizontal urban area (m2)
        * urbArea -- length of the side of the urban area orthogonal
        * orthLength -- length to side of urban area orthogonal to wind direction (m)
        * paralLength -- length of side of urban area parallel to wind direction (m)
        * ublTemp -- urban boundary layer temperature (K)
        * ublTempdx -- urban boundary layer temperature discretization (K)
        * dayBLHeight -- daytime mixing height, orig = 700
        * nightBLHeight -- nighttime boundary-layer ht (m), Sing: 80, Bub-Cap: 50, orig 80
    """

    def __init__(self, location, charLength, initialTemp, maxdx, dayBLHeight,
                 nightBLHeight):
        self.location = location
        self.charLength = charLength
        self.perimeter = 4. * charLength
        self.urbArea = charLength ** 2
        self.orthLength = charLength

        numdx = round(charLength / min(charLength, maxdx))
        self.paralLength = charLength / numdx
        self.ublTemp = initialTemp
        self.ublTempdx = [initialTemp for x in range(int(numdx))]
        self.dayBLHeight = dayBLHeight
        self.nightBLHeight = nightBLHeight

        # Logger will be disabled by default unless explicitly called in tests
        self.logger = logging.getLogger(__name__)

    def ublmodel(self, UCM, RSM, rural, forc, parameter, simTime):
        # Note that only one urban canyon area is considered
        self.sensHeat = UCM.sensHeat
        heatDif = max(self.sensHeat - rural.sens, 0)
        Cp = parameter.cp  # Heat capacity of air (J/kg.K)
        k_w = parameter.circCoeff  # k_w per Bueno 'the uwg', eq 8
        g = parameter.g  # Gravity
        v_wind = max(forc.wind, parameter.windMin)  # wind velocity

        # Air density
        refDens = 0.
        for iz in range(RSM.nzref):
            refDens = \
                refDens + RSM.densityProfC[iz] * RSM.dz[iz] / \
                (RSM.z[RSM.nzref - 1] + RSM.dz[RSM.nzref - 1] / 2.)

        forDens = 0.
        for iz in range(RSM.nzfor):
            forDens = \
                forDens + RSM.densityProfC[iz] * RSM.dz[iz] / \
                (RSM.z[RSM.nzfor - 1] + RSM.dz[RSM.nzfor - 1] / 2.)

        # ---------------------------------------------------------------------
        # Day
        # ---------------------------------------------------------------------
        time = simTime.secDay / 3600.
        noon = 12.
        daylimit = parameter.dayThreshold      # sunlight threshold for day (~150W/m^2)
        nightlimit = parameter.dayThreshold    # sunlight threshold for night (~50W/m^2)
        sunlight = forc.dir + forc.dif

        # If dir & dif light is greater than threshold, use day
        is_day = (sunlight > daylimit) and (time < noon or is_near_zero(time - noon)) \
            or (sunlight > nightlimit) and (time > noon) or (self.sensHeat > 150.0)
        if is_day:
            # Circulation velocity per Bueno 'the uwg', eq 8
            self.logger.debug("{} Day ubl calcs".format(__name__))
            h_UBL = self.dayBLHeight            # Day boundary layer height
            eqTemp = RSM.tempProf[RSM.nzref - 1]
            eqWind = RSM.windProf[RSM.nzref - 1]

            Csurf = UCM.Q_ubl * simTime.dt / (h_UBL * refDens * Cp)
            u_circ = k_w * (g * heatDif / Cp / refDens / eqTemp * h_UBL) ** (1. / 3.)

            if v_wind > u_circ:  # Forced problem (usually this)
                advCoef = self.orthLength*eqWind * simTime.dt / self.urbArea * 1.4
                self.ublTemp = (Csurf + advCoef * eqTemp + self.ublTemp) / (1. + advCoef)
                self.ublTempdx = [self.ublTemp for x in range(len(self.ublTempdx))]
            else:  # Convective problem
                advCoef = self.perimeter * u_circ * simTime.dt / self.urbArea * 1.4
                self.ublTemp = (Csurf + advCoef * eqTemp + self.ublTemp) / (1 + advCoef)
                self.ublTempdx = [self.ublTemp for x in range(len(self.ublTempdx))]

        # ---------------------------------------------------------------------
        # Night
        # ---------------------------------------------------------------------
        else:
            self.logger.debug("{} Night ubl calcs".format(__name__))
            h_UBL = self.nightBLHeight  # Night boundary layer height
            Csurf = UCM.Q_ubl * simTime.dt / (h_UBL * refDens * Cp)
            self.ublTemp, self.ublTempdx = \
                UBLDef.nightforc(self.ublTempdx, simTime.dt, h_UBL, self.paralLength,
                                 self.charLength, RSM, Csurf)

        self.logger.debug("ublTemp = {}".format(self.ublTemp))

    @staticmethod
    def nightforc(ublTempdx, dt, h_UBL, paralLength, charLength, RSM, Csurf):
        # Night forcing (RSM.nzfor = number of layers of forcing)
        # Average potential temperature & wind speed of the profile
        intAdv1 = 0.
        for iz in range(RSM.nzfor):
            intAdv1 = intAdv1 + RSM.windProf[iz] * RSM.tempProf[iz] * RSM.dz[iz]

        advCoef1 = 1.4 * dt / paralLength / h_UBL * intAdv1

        intAdv2 = 0
        for iz in range(RSM.nzfor):
            intAdv2 = intAdv2 + RSM.windProf[iz] * RSM.dz[iz]

        advCoef2 = 1.4 * dt / paralLength / h_UBL * intAdv2

        ublTempdx[0] = (Csurf + advCoef1 + ublTempdx[0]) / (1 + advCoef2)
        ublTemp = ublTempdx[0]

        for i in range(1, int(charLength) // int(paralLength)):
            eqTemp = ublTempdx[i - 1]
            ublTempdx[i] = (Csurf + advCoef2*eqTemp + ublTempdx[i]) / (1 + advCoef2)
            ublTemp = ublTemp + ublTempdx[i]

        ublTemp = ublTemp / float(charLength) * float(paralLength)

        return ublTemp, ublTempdx

    def __repr__(self):
        return 'UBL,\n location: {}\n charLength: {}\n perimeter: {}\n urbArea: {}\n ' \
            'orthLength: {}\n paralLength: {}\n ublTemp: {}\n dayBLHeight: {}\n ' \
            'nightBLHeight: {}'.format(
                self.location, self.charLength, self.perimeter, self.urbArea,
                self.orthLength, self.paralLength, self.ublTemp, self.dayBLHeight,
                self.nightBLHeight)



import pytest
from decimal import Decimal

def test_64bit_precision():
    """
    Tests for max number of precision for 64bit python float
    64 bits = 53 bits (significand) + 8 bits (characteristic)
    Therefore max fraction size for 53 bits in base 2 is (2^-53 = 1.11e-16)
    So for numbers > 0. we have to subtract the 16 significant digits (significand)
    from the integer part when testing the numbers
    """

    """
    # Unblock this for visual test
    print "0.12345678901234561111", "-- original input"
    print Decimal.from_float(0.12345678901234561111)
    print "  1234567890123456 -- will be accurate to 16th digit"
    print Decimal.from_float(11.12345678901234561111)
    print "   12345678901234 -- will be accurate to 14th digit"
    print Decimal.from_float(11111.12345678901234561111)
    print "      12345678901 -- will be accurate to 11th digit"
    print '----'
    """

    # 16 digits number
    digit_19 = 0.1234567890123456444
    trunc_at_11 = digit_19*10**11 - int(digit_19*10**11) # arbritrary truncation
    # trunc_at_11: 0.23456444
    #print Decimal.from_float(trunc_at_11)
    # != at 17+1 digits (give one additional digit for tolerance)
    assert trunc_at_11 != pytest.approx(0.23456444, abs=1e-7)
    # == at 16 digits
    assert trunc_at_11 == pytest.approx(0.23456444, abs=1e-5)

    # Equal at 11 digits
    digit_11 = 11111.12345678901444
    trunc_at_5 = digit_11*10**5 - int(digit_11*10**5) # arbritrary truncation
    # trunc_at_5: 0.678901444
    #print Decimal.from_float(trunc_at_5)
    # != at 17+1 digits (give one additional digit for tolerance)
    assert trunc_at_5 != pytest.approx(0.678901444, abs=1e-8)
    # == at 16 digits
    assert trunc_at_5 == pytest.approx(0.678901444, abs=1e-6)


if __name__ == "__main__":
    test_64bit_precision()



"""Test for uwg from df"""
from __future__ import division, print_function

try:
    import cPickle as pickle
except ImportError:
    import pickle

import os
import pytest
from .test_base import auto_setup_uwg


TEST_DIR = os.path.abspath(os.path.dirname(__file__))
EPW_PATH = os.path.join(TEST_DIR, 'epw', 'SGP_Singapore.486980_IWEC.epw')
PARAM_DIR = os.path.join(TEST_DIR, 'parameters')
PARAM_PATH = os.path.join(PARAM_DIR, 'initialize_fatal_error.uwg')
BUILDINGPARAMS = [
    'floor_height', 'int_heat', 'int_heat_night', 'int_heat_day', 'int_heat_frad',
    'int_heat_flat', 'infil', 'vent', 'glazing_ratio', 'u_value', 'shgc',
    'condtype', 'cop', 'cool_setpoint_day', 'cool_setpoint_night', 'heat_setpoint_day',
    'heat_setpoint_night', 'coolcap', 'heateff', 'msys', 'indoor_temp', 'indoor_hum',
    'heat_cap', 'cop_adj', 'canyon_fraction']


def check_obj_attr(obj1, obj2, attr_lst):
    for attr in attr_lst:
        v1, v2 = [getattr(obj, attr) for obj in [obj1, obj2]]

        # if v1 != v2:
        #     print('{}: {} != {}'.format(attr, v1, v2))
        #     is_equal = False
        # else:
        #     print('{}: {} == {}'.format(attr, v1, v2))

        if isinstance(v1, (int, float)) and isinstance(v2, (int, float)):
            assert v1 == pytest.approx(v2, abs=1e-10)
        else:
            assert v1 == v2


def unpickle(cpickle_path):
    with open(cpickle_path, 'rb') as f:
        try:
            return pickle.load(f)
        except EOFError:
            return None


def uwg_from_df():

    # spelling mistake
    # lathAnth = in DF = latAnth
    testuwg = auto_setup_uwg(EPW_PATH, PARAM_PATH)
    # Error on Aug 22 if dt at 300
    testuwg.month = 8
    testuwg.day = 1
    testuwg.nday = 31

    bem = unpickle(os.path.join(PARAM_DIR, 'initialize_fatal_error_bem.pkl'))
    testuwg._compute_BEM()
    testuwg.BEM = bem

    testuwg._read_epw()
    testuwg._compute_input()
    testuwg._hvac_autosize()

    return testuwg


def uwg_manual():
    # spelling mistake
    # lathAnth = in DF = latAnth
    testuwg = auto_setup_uwg(EPW_PATH, PARAM_PATH)
    testuwg._compute_BEM()

    # # Add custom typology parameters from DF
    df_typology_params = [[4.0, 0.5, 0.65, 0.35, 0.08, 0.5, 0.5],
                          [3.05, 0.5, 0.1499, 0.54, 0.15, 0.2, 0.2],
                          [4.0, 0.5, 0.0058, 0.251, 0.08, 0.2, 0.2]]

    # lathAnth type BEMDef:
    # Type = LargeOffice, Zone = 1A (Miami), Era = Pst80, Construction = MassWall
    # type BEMDef:
    #   Type = MidRiseApartment, Zone = 1A (Miami), Era = Pre80, Construction=SteelFrame
    # type BEMDef:
    #   Type = WareHouse, Zone = 1A (Miami), Era = Pst80, Construction = MassWall

    for i in range(len(testuwg.BEM)):
        testuwg.BEM[i].building.floor_height = df_typology_params[i][0]
        testuwg.BEM[i].building.canyon_fraction = df_typology_params[i][1]
        testuwg.BEM[i].building.glazing_ratio = df_typology_params[i][2]
        testuwg.BEM[i].building.shgc = df_typology_params[i][3]
        testuwg.BEM[i].wall.albedo = df_typology_params[i][4]
        testuwg.BEM[i].roof.albedo = df_typology_params[i][5]
        testuwg.BEM[i].roof.vegcoverage = df_typology_params[i][6]

    # Error on Aug 22 if dt at 300
    testuwg.month = 8
    testuwg.day = 1
    testuwg.nday = 31

    testuwg._read_epw()
    testuwg._compute_input()
    testuwg._hvac_autosize()

    return testuwg


# TODO: Repickle once variable names have been finished PEP8 conversion
# def test_compare():
#     uwgpkl = uwg_from_df()
#     uwgfile = uwg_manual()

#     for i in range(3):
#         check_obj_attr(uwgpkl.BEM[i].building, uwgfile.BEM[i].building, BUILDINGPARAMS)


def test_fatal_error_manual():
    # These will fail if simTime.dt at 300 o Aug 22nd
    # Will raise the FATAL ERROR exception
    with pytest.raises(Exception):
        uwgpkl = uwg_manual()
        uwgpkl.simulate()


def test_fatal_error_df():
    # These will fail if simTime.dt at 300 o Aug 22nd
    # Will raise the FATAL ERROR exception
    with pytest.raises(Exception):
        uwgpkl = uwg_from_df()
        uwgpkl.simulate()



"""Test cli result module."""
from click.testing import CliRunner
from uwg.cli.simulate import simulate
from uwg.cli.validate import validate
from uwg.cli import viz

import os
import sys


def test_viz():
    runner = CliRunner()
    result = runner.invoke(viz)
    assert result.exit_code == 0, result.output
    assert result.output.startswith('vi'), result.output
    assert result.output.endswith('z!\n'), result.output


def test_model_validate():
    """Test uwg validation."""

    if (sys.version_info >= (3, 7)):
        runner = CliRunner()
        input_uwg = './tests/json/uwg.json'
        result = runner.invoke(validate, ['model', input_uwg])
        assert result.exit_code == 0, result.output

        input_uwg = './tests/json/custom_uwg.json'
        result = runner.invoke(validate, ['model', input_uwg])
        assert result.exit_code == 0, result.output


def test_param_validate():
    """Test uwg validation."""
    if (sys.version_info >= (3, 7)):
        runner = CliRunner()
        input_uwg = './tests/parameters/initialize_singapore.uwg'
        result = runner.invoke(validate, ['param', input_uwg])
        print(result.output)
        assert result.exit_code == 0, result.output

        input_uwg = './tests/parameters/initialize_singapore_error.uwg'
        result = runner.invoke(validate, ['param', input_uwg])
        assert result.exit_code == 1, result.output


def test_uwg_simulate():
    """Test simple uwg simulation."""

    runner = CliRunner()

    # test with basic command
    input_uwg = './tests/json/uwg.json'
    epw_path = './tests/epw/SGP_Singapore.486980_IWEC.epw'

    result = runner.invoke(simulate, ['model', input_uwg, epw_path])
    assert result.exit_code == 0, result.output

    output_epw = './tests/epw/SGP_Singapore.486980_IWEC_UWG.epw'
    assert os.path.isfile(output_epw)
    os.remove(output_epw)

    # test with optional arguments
    new_epw_dir = './tests/epw_uwg'
    new_epw_name = 'custom_test_singapore.epw'
    result = runner.invoke(
        simulate, ['model', input_uwg, epw_path, '--new-epw-dir',
                   new_epw_dir, '--new-epw-name', new_epw_name])
    assert result.exit_code == 0, result.output

    default_output_epw = './tests/epw/SGP_Singapore.486980_IWEC_UWG.epw'
    assert not os.path.isfile(default_output_epw)

    output_epw = os.path.join(new_epw_dir, new_epw_name)
    assert os.path.isfile(output_epw)
    os.remove(output_epw)


def test_custom_uwg_simulate():
    """Test custom uwg simulation."""

    runner = CliRunner()

    # test with basic command
    input_uwg = './tests/json/custom_uwg.json'
    epw_path = './tests/epw/SGP_Singapore.486980_IWEC.epw'
    result = runner.invoke(
        simulate, ['model', input_uwg, epw_path])
    assert result.exit_code == 0, result.output

    output_epw = './tests/epw/SGP_Singapore.486980_IWEC_UWG.epw'
    assert os.path.isfile(output_epw)
    os.remove(output_epw)


def test_param_simulate():
    """Test simple uwg simulation from param file."""

    runner = CliRunner()

    # test with basic command
    input_uwg = './tests/parameters/initialize_singapore.uwg'
    epw_path = './tests/epw/SGP_Singapore.486980_IWEC.epw'

    result = runner.invoke(simulate, ['param', input_uwg, epw_path])
    assert result.exit_code == 0, result.output

    output_epw = './tests/epw/SGP_Singapore.486980_IWEC_UWG.epw'
    assert os.path.isfile(output_epw)
    os.remove(output_epw)



"""Class to estimate deep ground/water temperature."""
from __future__ import division


class Forcing (object):
    """Force method to estimate deep ground/water temperature.

    Args:
        staTemp: List of hourly air temperature for simulation period.
        weather: Weather object.

    Properties:
        * deepTemp -- deep soil temperature (K)
        * waterTemp -- ground water temp, set to temp at 2m
        * infra -- horizontal Infrared Radiation Intensity (W m-2)
        * uDir -- wind direction
        * hum -- specific humidty (kg kg-1)
        * pres -- Pressure (Pa)
        * temp -- air temperature (C)
        * rHum -- Relative humidity (%)
        * dir -- normal solar direct radiation (W m-2)
        * dif -- horizontal solar diffuse radiation (W m-2)
        * prec -- precipitation (mm h-1)
        * wind -- wind speed (m s-1)
    """

    def __init__(self, staTemp=None, weather=None):

        if not (staTemp and weather):
            # Define default values for instance variables
            self.deepTemp = None
            self.waterTemp = None
            self.infra = None
            self.uDir = None
            self.hum = None
            self.pres = None
            self.temp = None
            self.rHum = None
            self.dir = None
            # dif: horizontal solar diffuse radiation (W m-2)
            # ...Amount of solar radiation received from the sky
            # ...(excluding the solar disk) on a horizontal surface
            self.dif = None
            self.prec = None
            self.wind = None
        else:
            self.deepTemp = sum(staTemp) / float(len(staTemp))
            self.waterTemp = sum(staTemp) / float(len(staTemp))
            self.infra = weather.staInfra
            self.uDir = weather.staUdir
            self.hum = weather.staHum
            self.pres = weather.staPres
            self.temp = weather.staTemp
            self.rHum = weather.staRhum
            self.dir = weather.staDir
            self.dif = weather.staDif
            self.prec = [p / 3.6e6 for p in weather.staRobs]
            self.wind = weather.staUmod

    def __repr__(self):
        return 'forcing,\n deepTemp: {}\n waterTemp: {}'.format(
            self.deepTemp, self.waterTemp)




"""UWG JSON model validation commands."""
import sys
import os
import json
import logging
try:
    import click
except ImportError:
    raise ImportError(
        'click is not installed. Try `pip install . [cli]` command.'
    )

from uwg import UWG

try:
    import uwg_schema.model as schema_model
except ImportError:
    pass  # uwg_schema is not installed

_logger = logging.getLogger(__name__)


@click.group(help='Commands for validating UWG JSON and .uwg files.')
def validate():
    pass


@validate.command('model')
@click.argument('model-json')
def validate_model(model_json):
    """Validate a UWG model JSON file against the UWG schema.

    \b
    Args:
        model_json: Full path to a UWG model JSON file. Note that this will
            not check if rural and new .epw file paths are valid since these
            can be overridden by CLI arguments.
    """
    try:
        try:
            schema_model
        except Exception:
            raise ImportError(
                'uwg_schema is not installed. Try `pip install . [cli]` command.')
        assert os.path.isfile(
            model_json), 'No JSON file found at {}.'.format(model_json)

        # validate the Model JSON
        click.echo('Validating Model JSON ...')
        schema_model.UWG.parse_file(model_json)
        click.echo('Pydantic validation passed.')
        with open(model_json) as json_file:
            data = json.load(json_file)
        UWG.from_dict(data)
        click.echo('Python re-serialization passed.')
        click.echo('Congratulations! Your UWG model JSON is valid!')
    except Exception as e:
        _logger.exception('Model validation failed.\n{}'.format(e))
        sys.exit(1)
    else:
        sys.exit(0)


@validate.command('param')
@click.argument('param-uwg')
def validate_param(param_uwg):
    """Validate a .uwg parameter file.

    \b
    Args:
        param_uwg: Full path to a .uwg parameter file.
    """
    try:
        try:
            schema_model
        except Exception:
            raise ImportError(
                'uwg_schema is not installed. Try `pip install . [cli]` command.')

        # validate the Model JSON
        click.echo('Validating .uwg parameter file ...')
        UWG.from_param_file(param_uwg)
        click.echo('Congratulations! Your .uwg parameter file is valid!')
    except Exception as e:
        _logger.exception('The .uwg parameter file validation failed.\n{}'.format(e))
        sys.exit(1)
    else:
        sys.exit(0)



import cPickle
import os

def get_bem_pickle(uwg_data, refdir, fname):

    # Pickle objects, protocol 1 b/c binary file
    pkl_file_path = os.path.join(refdir,fname+".pkl")
    with open(pkl_file_path, "wb") as outf:
        cPickle.dump(uwg_data.BEM, outf)

    return pkl_file_path

def get_uwg_file(uwg_object, refdir, fname):
    uwg_file_path = os.path.join(refdir,fname+".uwg")
    f = open(uwg_file_path, "w")

    f.write("# =================================================\n")
    f.write("# REQUIRED PARAMETERS\n")
    f.write("# =================================================\n")
    f.write("\n")
    f.write("# Urban characteristics\n")
    f.write("bldHeight,{},\n".format(uwg_object.bldHeight))
    f.write("bldDensity,{},\n".format(uwg_object.bldDensity))
    f.write("verToHor,{},\n".format(uwg_object.verToHor))
    f.write("h_mix,{},\n".format(uwg_object.h_mix))
    f.write("charLength,{},\n".format(uwg_object.charLength))  # dimension of a square that encompasses the whole neighborhood [aka. characteristic length] (m)
    f.write("albRoad,{},\n".format(uwg_object.alb_road))      # road albedo (0 - 1)
    f.write("dRoad,{},\n".format(uwg_object.d_road))        # road pavement thickness (m)
    f.write("kRoad,{},\n".format(uwg_object.kRoad))         # road pavement conductivity (W/m K)
    f.write("cRoad,{},\n".format(uwg_object.cRoad))    # road volumetric heat capacity (J/m^3 K)
    f.write("sensAnth,{},\n".format(uwg_object.sensAnth))      # non-building sensible heat at street level [aka. heat from cars, pedestrians, street cooking, etc. ] (W/m^2)
    f.write("latAnth,{},\n".format(uwg_object.latAnth))        # non-building latent heat (W/m^2) (currently not used)
    f.write("\n")
    f.write("zone,{},\n".format(uwg_object.zone + 1))
    f.write("\n")
    f.write("# Vegetation parameters\n")
    f.write("vegCover,{},\n".format(uwg_object.vegCover))     # Fraction of the urban ground covered in grass/shrubs only (0-1)
    f.write("treeCoverage,{},\n".format(uwg_object.treeCoverage)) # Fraction of the urban ground covered in trees (0-1)
    f.write("vegStart,{},\n".format(uwg_object.vegStart))       # The month in which vegetation starts to evapotranspire (leaves are out)
    f.write("vegEnd,{},\n".format(uwg_object.vegEnd))        # The month in which vegetation stops evapotranspiring (leaves fall)
    f.write("albVeg,{},\n".format(uwg_object.albVeg))      # Vegetation albedo
    f.write("rurVegCover,{},\n".format(uwg_object.rurVegCover))  # Fraction of the rural ground covered by vegetation
    f.write("latGrss,{},\n".format(uwg_object.latGrss))      # Fraction of the heat absorbed by grass that is latent. Used in UWG only to calculate sensible heat fraction.
    f.write("latTree,{},\n".format(uwg_object.latTree))      # Fraction of the heat absorbed by trees that is latent. Used in UWG only to calculate sensible heat fraction.
    f.write("\n")
    f.write("# Traffic schedule [1 to 24 hour],\n")
    f.write("SchTraffic,\n")
    for i in range(3):
        for j in range(24):
            f.write("{},".format(uwg_object.SchTraffic[i][j]))
        f.write("\n")
    f.write("\n")
    f.write("# Fraction of building stock for each DOE Building type (pre-80's build, 80's-present build, new)\n")
    f.write("# Note that sum(bld) must be equal to 1\n")
    f.write("bld,\n")
    for i in range(16):
        for j in range(3):
            f.write("{},".format(uwg_object.bld[i][j]))
        f.write("\n")
    f.write("\n")
    f.write("# =================================================\n")
    f.write("# OPTIONAL URBAN PARAMETERS\n")
    f.write("# =================================================\n")
    f.write("# If not provided, optional parameters are taken from corresponding DOE Reference building\n")
    f.write("albRoof,{},\n".format(uwg_object.albRoof if uwg_object.albRoof else ""))  # roof albedo (0 - 1)
    f.write("vegRoof,{},\n".format(uwg_object.vegRoof if uwg_object.vegRoof else ""))  # Fraction of the roofs covered in grass/shrubs (0 - 1)
    f.write("glzR,{},\n".format(uwg_object.glzR if uwg_object.glzR else ""))     # Glazing Ratio (0 - 1)
    f.write("SHGC,{},\n".format(uwg_object.SHGC if uwg_object.SHGC else ""))     # Solar Heat Gain Coefficient (0 - 1)
    f.write("albWall,{},\n".format(uwg_object.albWall if uwg_object.albWall else ""))  # wall albedo (0 - 1)
    f.write("\n")
    f.write("# =================================================\n")
    f.write("# OPTIONAL PARAMETERS FOR SIMULATION CONTROL,\n")
    f.write("# =================================================\n")
    f.write("\n")
    f.write("# Simulation parameters,\n")
    f.write("Month,{},\n".format(uwg_object.Month))        # starting month (1-12)
    f.write("Day,{},\n".format(uwg_object.Day))          # starting day (1-31)
    f.write("nDay,{},\n".format(uwg_object.nDay))        # number of days to run simultion
    f.write("dtSim,{},\n".format(uwg_object.dtSim))      # simulation time step (s)
    f.write("dtWeather,{},\n".format(uwg_object.dtWeather)), # weather time step (s)
    f.write("\n")
    f.write("# HVAC system and internal loads\n")
    f.write("autosize,{},\n".format(uwg_object.autosize))     # autosize HVAC (1 for yes; 0 for no)
    f.write("sensOcc,{},\n".format(uwg_object.sensOcc))    # Sensible heat per occupant (W)
    f.write("LatFOcc,{},\n".format(uwg_object.LatFOcc))    # Latent heat fraction from occupant (normally 0.3)
    f.write("RadFOcc,{},\n".format(uwg_object.RadFOcc))    # Radiant heat fraction from occupant (normally 0.2)
    f.write("RadFEquip,{},\n".format(uwg_object.RadFEquip))  # Radiant heat fraction from equipment (normally 0.5)
    f.write("RadFLight,{},\n".format(uwg_object.RadFLight))  # Radiant heat fraction from light (normally 0.7)
    f.write("\n")
    f.write("#Urban climate parameters\n")
    f.write("h_ubl1,{},\n".format(uwg_object.h_ubl1))    # ubl height - day (m)
    f.write("h_ubl2,{},\n".format(uwg_object.h_ubl2))      # ubl height - night (m)
    f.write("h_ref,{},\n".format(uwg_object.h_ref))      # inversion height (m)
    f.write("h_temp,{},\n".format(uwg_object.h_temp))       # temperature height (m)
    f.write("h_wind,{},\n".format(uwg_object.h_wind))     # wind height (m)
    f.write("c_circ,{},\n".format(uwg_object.c_circ))   # circulation coefficient (default = 1.2 per Bruno (2012))
    f.write("c_exch,{},\n".format(uwg_object.c_exch))      # exchange coefficient (default = 1; ref Bruno (2014))
    f.write("maxDay,{},\n".format(uwg_object.maxDay))     # max day threshold (W/m^2)
    f.write("maxNight,{},\n".format(uwg_object.maxNight))    # max night threshold (W/m^2)
    f.write("windMin,{},\n".format(uwg_object.windMin))      # min wind speed (m/s)
    f.write("h_obs,{},\n".format(uwg_object.h_obs))      # rural average obstacle height (m)

    f.close()

    return uwg_file_path

if x and y and z and run:

    _uwg_object = x
    _dir = y
    _fname = z

    print "Remember not to run_ the 'RunUWG' component. Just write the file!"
    uwg_file = get_uwg_file(_uwg_object, _dir, _fname)
    bem_file = get_bem_pickle(_uwg_object, _dir, _fname)

    fout = [uwg_file, bem_file]

    print fout



"""Test for Weather class."""

import pytest
from uwg import SimParam, Weather
from .test_base import DEFAULT_EPW_PATH


def test_weather():
    """Test for weather.py"""

    dtSim = 300             # Sim time step
    dtWeather = 3600        # Weather data time-step
    MONTH = 7               # Begin month
    DAY = 30                # Begin day of the month
    NUM_DAYS = 7            # Number of days of simulation

    simTime = SimParam(dtSim, dtWeather, MONTH, DAY, NUM_DAYS)
    weather = Weather(DEFAULT_EPW_PATH, simTime.timeInitial, simTime.timeFinal)

    # Weather Tests
    assert len(weather.staDif) == \
        pytest.approx(simTime.timeFinal - simTime.timeInitial + 1, abs=1e-6)
    assert len(weather.staHum) == \
        pytest.approx(simTime.timeFinal - simTime.timeInitial + 1, abs=1e-6)
    assert len(weather.staTemp) == \
        pytest.approx(simTime.timeFinal - simTime.timeInitial + 1, abs=1e-6)
    assert weather.staTemp[3] == pytest.approx(24.+273.15, abs=1e-6)
    assert weather.staTemp[-1] == pytest.approx(27.+273.15, abs=1e-6)
    assert weather.staUdir[2] == pytest.approx(270, abs=1e-1)  # 270 deg
    assert weather.staUmod[4] == pytest.approx(.5, abs=1e-6)   # 0.5 m/s
    assert weather.staPres[10] == pytest.approx(100600., abs=1e-1)
    assert weather.staInfra[13] == pytest.approx(428., abs=1e-1)
    assert weather.staDif[6] == pytest.approx(0., abs=1e-3)
    assert weather.staDif[8] == pytest.approx(95., abs=1e-6)
    assert weather.staRobs[8] == pytest.approx(0.0, abs=1e-3)  # 0. mm/hre



"""uwg simulation running commands."""
import sys
import json
import logging
try:
    import click
except ImportError:
    raise ImportError(
        'click is not installed. Try `pip install . [cli]` command.'
    )

from uwg import UWG

_logger = logging.getLogger(__name__)


@click.group(help='Commands for simulating UWG models.')
def simulate():
    pass


@simulate.command('model')
@click.argument('model-json', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.argument('epw-path', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.option('--new-epw-dir', help='Optional argument for the destination directory '
              'into which the morphed .epw file is written. The argument passed here '
              'will overwrite the new_epw_dir specified in the UWG JSON model file.',
              default=None, show_default=True)
@click.option('--new-epw-name', help='Optional argument for The destination file name '
              'of the morphed .epw file. The argument passed here will overwrite the '
              'new_epw_name specified in the UWG JSON model file.',
              default=None, show_default=True)
def simulate_json_model(model_json, epw_path, new_epw_dir, new_epw_name):
    """Simulate a UWG model from a JSON model file.

    \b
    Args:
        model_json: Full path to a JSON model file.
        epw_path: Full path of the rural .epw file that will be morphed.
    """
    try:
        with open(model_json) as json_file:
            data = json.load(json_file)

        uwg_model = UWG.from_dict(
            data, epw_path=epw_path, new_epw_dir=new_epw_dir, new_epw_name=new_epw_name)
        uwg_model.generate()
        uwg_model.simulate()
        uwg_model.write_epw()
    except Exception as e:
        _logger.exception('UWG model simulation failed.\n{}'.format(e))
        sys.exit(1)
    else:
        sys.exit(0)


@simulate.command('param')
@click.argument('param-uwg', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.argument('epw-path',  type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.option('--new-epw-dir', help='Optional argument for the destination directory '
              'into which the morphed .epw file is written.',
              default=None, show_default=True)
@click.option('--new-epw-name', help='Optional argument for The destination file name '
              'of the morphed .epw file.',
              default=None, show_default=True)
def simulate_uwg_param_model(param_uwg, epw_path, new_epw_dir, new_epw_name):
    """Simulate a UWG model from a .uwg parameter file.

    \b
    Args:
        param_uwg: Full path to a .uwg param file.
        epw_path: Full path of the rural .epw file that will be morphed.
    """
    try:
        uwg_model = UWG.from_param_file(
            param_uwg, epw_path, new_epw_dir, new_epw_name)
        uwg_model.generate()
        uwg_model.simulate()
        uwg_model.write_epw()
    except Exception as e:
        _logger.exception('UWG model simulation failed.\n{}'.format(e))
        sys.exit(1)
    else:
        sys.exit(0)
